import sqlite3
from pathlib import Path
import pandas as pd
from py_neuromodulation.utils.types import _PathLike
from py_neuromodulation.utils.io import generate_unique_filename


class NMDatabase:
    """
    Class to create a database and insert data into it.
    Parameters
    ----------
    out_dir : _PathLike
        The directory to save the database.
    csv_path : str, optional
        The path to save the csv file. If not provided, it will be saved in the same folder as the database.
    """

    def __init__(
        self,
        name: str,
        out_dir: _PathLike,
        csv_path: _PathLike | None = None,
    ):
        # Make sure out_dir exists
        Path(out_dir).mkdir(parents=True, exist_ok=True)

        self.db_path = Path(out_dir, f"{name}.db")

        self.table_name = f"{name}_data"  # change to param?
        self.table_created = False

        if self.db_path.exists():
            self.db_path = generate_unique_filename(self.db_path)
            name = self.db_path.stem

        if csv_path is None:
            self.csv_path = Path(out_dir, f"{name}.csv")
        else:
            self.csv_path = Path(csv_path)

        self.csv_path.parent.mkdir(parents=True, exist_ok=True)

        self.conn = sqlite3.connect(self.db_path)
        self.cursor = self.conn.cursor()

        # Database config and optimization, prioritize data integrity
        self.cursor.execute("PRAGMA journal_mode=WAL")  # Write-Ahead Logging mode
        self.cursor.execute("PRAGMA synchronous=FULL")  # Sync on every commit
        self.cursor.execute("PRAGMA temp_store=MEMORY")  # Store temp tables in memory
        self.cursor.execute(
            "PRAGMA wal_autocheckpoint = 1000"
        )  # WAL checkpoint every 1000 pages (default, 4MB, might change)
        self.cursor.execute(
            f"PRAGMA mmap_size = {2 * 1024 * 1024 * 1024}"
        )  # 2GB of memory mapped

    def infer_type(self, value):
        """Infer the type of the value to create the table schema.
        Parameters
        ----------
        value : int, float, str
            The value to infer the type."""

        if isinstance(value, (int, float)):
            return "REAL"
        elif isinstance(value, str):
            return "TEXT"
        else:
            return "BLOB"

    def create_table(self, feature_dict: dict):
        """
        Create a table in the database.
        Parameters
        ----------
        feature_dict : dict
            The dictionary with the feature names and values.
        """
        columns_schema = ", ".join(
            [
                f'"{column}" {self.infer_type(value)}'
                for column, value in feature_dict.items()
            ]
        )

        self.cursor.execute(
            f'CREATE TABLE IF NOT EXISTS "{self.table_name}" ({columns_schema})'
        )

        # Create column names and placeholders for insert statement
        self.columns: str = ", ".join([f'"{column}"' for column in feature_dict.keys()])
        # Use named placeholders for more resiliency against unexpected change in column order
        self.placeholders = ", ".join([f":{key}" for key in feature_dict.keys()])

    def insert_data(self, feature_dict: dict):
        """
        Insert data into the database.
        Parameters
        ----------
        feature_dict : dict
            The dictionary with the feature names and values.
        """

        if not self.table_created:
            self.create_table(feature_dict)
            self.table_created = True

        insert_sql = f'INSERT INTO "{self.table_name}" ({self.columns}) VALUES ({self.placeholders})'

        self.cursor.execute(insert_sql, feature_dict)

    def commit(self):
        self.conn.commit()

    def fetch_all(self):
        """ "
        Fetch all the data from the database.
        Returns
        -------
        pd.DataFrame
            The data in a pandas DataFrame.
        """
        return pd.read_sql_query(f'SELECT * FROM "{self.table_name}"', self.conn)

    def head(self, n: int = 5):
        """ "
        Returns the first N rows of the database.
        Parameters
        ----------
        n : int, optional
            The number of rows to fetch, by default 1
        -------
        pd.DataFrame
            The data in a pandas DataFrame.
        """
        return pd.read_sql_query(
            f'SELECT * FROM "{self.table_name}" LIMIT {n}', self.conn
        )

    def save_as_csv(self):
        df = self.fetch_all()
        df.to_csv(self.csv_path, index=False)

    def close(self):
        # Optimize before closing is recommended:
        # https://www.sqlite.org/pragma.html#pragma_optimize
        self.cursor.execute("PRAGMA optimize")
        self.conn.close()



from py_neuromodulation.utils import database
import pandas as pd


def test_db_setup():
    nm_db = database.NMDatabase(
        name="test_db",
        out_dir="test_data",
    )

    assert nm_db.db_path.exists()

    features = {
        "feature_1": 1,
        "feature_2": 2,
        "feature_3": 3,
    }
    nm_db.create_table(features)

    # read table
    table = nm_db.cursor.execute(f"SELECT * FROM {nm_db.table_name}")

    assert table.fetchone() is None

    assert [info[0] for info in table.description] == [
        "feature_1",
        "feature_2",
        "feature_3",
    ]

    nm_db.insert_data(features)

    entry = nm_db.head(1)

    nm_db.save_as_csv()

    assert nm_db.csv_path.exists()
    assert pd.read_csv(nm_db.csv_path).equals(entry)



import multiprocessing as mp
import logging
from typing import Sequence
import sys
from pathlib import Path
from py_neuromodulation.utils.types import _PathLike
from functools import lru_cache
import platform
from py_neuromodulation import logger


def force_terminate_process(
    process: mp.Process, name: str, logger: logging.Logger | None = None
) -> None:
    log = logger.debug if logger else print

    import psutil

    p = psutil.Process(process.pid)
    try:
        log(f"Terminating process {name}")
        for child in p.children(recursive=True):
            log(f"Terminating child process {child.pid}")
            child.terminate()
        p.terminate()
        p.wait(timeout=3)
    except psutil.NoSuchProcess:
        log(f"Process {name} has already exited.")
    except psutil.TimeoutExpired:
        log(f"Forcefully killing {name}...")
        p.kill()


def create_logger(name, color: str, level=logging.INFO):
    """Function to set up a logger with color coded output"""
    color = ansi_color(color=color, bright=True, styles=["BOLD"])
    logger = logging.getLogger(name)
    log_format = f"{color}[%(name)s %(levelname)s (%(asctime)s)]:{ansi_color(styles=['RESET'])} %(message)s"
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(log_format, "%H:%M:%S"))
    stream_handler.setStream(sys.stderr)
    logger.setLevel(level)
    logger.addHandler(stream_handler)

    return logger


def ansi_color(
    color: str = "DEFAULT",
    bright: bool = True,
    styles: Sequence[str] = [],
    bg_color: str = "DEFAULT",
    bg_bright: bool = True,
) -> str:
    """
    Function to generate ANSI color codes for colored text in the terminal.
    See https://en.wikipedia.org/wiki/ANSI_escape_code

    Returns:
        str: ANSI color code
    """
    ANSI_COLORS = {
        # https://en.wikipedia.org/wiki/ANSI_escape_code
        "BLACK": 30,
        "RED": 31,
        "GREEN": 32,
        "YELLOW": 33,
        "BLUE": 34,
        "MAGENTA": 35,
        "CYAN": 36,
        "WHITE": 37,
        "DEFAULT": 39,
    }

    ANSI_STYLES = {
        "RESET": 0,
        "BOLD": 1,
        "FAINT": 2,
        "ITALIC": 3,
        "UNDERLINE": 4,
        "BLINK": 5,
        "NEGATIVE": 7,
        "CROSSED": 9,
    }

    color = color.upper()
    bg_color = bg_color.upper()
    styles = [style.upper() for style in styles]

    if color not in ANSI_COLORS.keys() or bg_color not in ANSI_COLORS.keys():
        raise ValueError(f"Invalid color: {color}")

    for style in styles:
        if style not in ANSI_STYLES.keys():
            raise ValueError(f"Invalid style: {style}")

    color_code = str(ANSI_COLORS[color] + (60 if bright else 0))
    bg_color_code = str(ANSI_COLORS[bg_color] + 10 + (60 if bg_bright else 0))
    style_codes = ";".join((str(ANSI_STYLES[style]) for style in styles))

    return f"\033[{style_codes};{color_code};{bg_color_code}m"


ansi_reset = ansi_color(styles=["RESET"])


def is_hidden(filepath: _PathLike) -> bool:
    """Check if a file or directory is hidden.

    Args:
        filepath (str): Path to the file or directory.

    Returns:
        bool: True if the file or directory is hidden, False otherwise.
    """
    from pathlib import Path

    filepath = Path(filepath)

    if sys.platform.startswith("win"):
        import ctypes

        try:
            attrs = ctypes.windll.kernel32.GetFileAttributesW(str(filepath))
            assert attrs != -1
            result = bool(attrs & 2) or filepath.name.startswith(".")
        except (AttributeError, AssertionError):
            result = filepath.name.startswith(".")
    else:
        result = filepath.name.startswith(".")

    return result


@lru_cache(maxsize=1)
def get_quick_access():
    system = platform.system()
    if system == "Windows":
        return get_windows_quick_access()
    elif system == "Darwin":  # macOS
        return get_macos_quick_access()
    else:  # Linux, Unix, etc.
        return {"items": []}


def get_windows_quick_access():
    quick_access_items = []

    # Add available drives
    available_drives = [
        f"{d}:\\" for d in "ABCDEFGHIJKLMNOPQRSTUVWXYZ" if Path(f"{d}:").exists()
    ]
    for drive in available_drives:
        quick_access_items.append(
            {"name": f"Drive ({drive})", "type": "drive", "path": drive}
        )

    # Get user's pinned folders
    pinned_folders = get_pinned_folders_windows()
    for folder in pinned_folders:
        path = Path(folder["Path"])
        if path.exists():
            quick_access_items.append(
                {"name": folder["Name"], "type": "folder", "path": str(path)}
            )

    # Get user's home directory
    home_path = Path.home()

    # Add common folders if they're not already in pinned folders
    common_folders = [
        ("Desktop", "Desktop"),
        ("Documents", "Documents"),
        ("Downloads", "Downloads"),
        ("Pictures", "Pictures"),
        ("Music", "Music"),
        ("Videos", "Videos"),
    ]

    for folder_name, folder_path in common_folders:
        full_path = home_path / folder_path
        if full_path.exists() and str(full_path) not in [
            item["path"] for item in quick_access_items
        ]:
            quick_access_items.append(
                {"name": folder_name, "type": "folder", "path": str(full_path)}
            )

    # Add user's home directory if not already included
    if str(home_path) not in [item["path"] for item in quick_access_items]:
        quick_access_items.append(
            {"name": "Home", "type": "folder", "path": str(home_path)}
        )

    return {"items": quick_access_items}


def get_pinned_folders_windows():
    import subprocess
    import json

    powershell_command = """
    $shell = New-Object -ComObject Shell.Application
    $quickaccess = $shell.Namespace("shell:::{679f85cb-0220-4080-b29b-5540cc05aab6}").Items()
    $pinned = $quickaccess | Where-Object { $_.IsFolder } | ForEach-Object {
        [PSCustomObject]@{
            Name = $_.Name
            Path = $_.Path
        }
    }
    $pinned | ConvertTo-Json
    """

    try:
        result = subprocess.run(
            ["powershell", "-Command", powershell_command],
            capture_output=True,
            text=True,
            check=True,
        )
        return json.loads(result.stdout)
    except subprocess.CalledProcessError as e:
        logger.error(f"Error running PowerShell command: {e}")
        return []
    except json.JSONDecodeError as e:
        logger.error(f"Error decoding JSON: {e}")
        return []


def get_macos_quick_access():
    quick_access_folders = get_macos_favorites()
    quick_access_items = []

    quick_access_items.append({"name": "Computer", "type": "drive", "path": "/"})

    # Add Volumes for macOS
    volumes_path = Path("/Volumes")
    if volumes_path.exists():
        for volume in volumes_path.iterdir():
            if volume.is_mount():
                quick_access_items.append(
                    {"name": volume.name, "type": "drive", "path": str(volume)}
                )

    # Add quick access folders
    for folder in quick_access_folders:
        path = Path(folder["Path"])
        if path.exists():
            quick_access_items.append(
                {"name": folder["Name"], "type": "folder", "path": str(path)}
            )

    # Add user's home directory if not already included
    home_path = str(Path.home())
    if home_path not in [item["path"] for item in quick_access_items]:
        quick_access_items.append({"name": "Home", "type": "folder", "path": home_path})

    return {"items": quick_access_items}


def get_macos_favorites():
    import subprocess
    import json

    favorites = []

    try:
        # Common locations in macOS
        common_locations = [
            ("Desktop", Path.home() / "Desktop"),
            ("Documents", Path.home() / "Documents"),
            ("Downloads", Path.home() / "Downloads"),
            ("Pictures", Path.home() / "Pictures"),
            ("Music", Path.home() / "Music"),
            ("Movies", Path.home() / "Movies"),
        ]

        for name, path in common_locations:
            if path.exists():
                favorites.append({"Name": name, "Path": str(path)})

        # Get user-defined favorites from sidebar plist
        plist_path = (
            Path.home()
            / "Library/Application Support/com.apple.sharedfilelist/com.apple.LSSharedFileList.FavoriteItems.sfl2"
        )
        if plist_path.exists():
            try:
                result = subprocess.run(
                    ["plutil", "-convert", "json", "-o", "-", str(plist_path)],
                    capture_output=True,
                    text=True,
                    check=True,
                )
                plist_data = json.loads(result.stdout)
                for item in plist_data.get("Bookmark", []):
                    if "Name" in item and "URL" in item:
                        path = item["URL"].replace("file://", "")
                        favorites.append({"Name": item["Name"], "Path": path})
            except (subprocess.CalledProcessError, json.JSONDecodeError) as e:
                logger.error(f"Error processing macOS favorites: {e}")

    except Exception as e:
        logger.error(f"Error getting macOS favorites: {e}")

    return favorites



import multiprocessing as mp
import threading
import os
import signal
import time
import platform

import logging

from .app_utils import force_terminate_process, create_logger, ansi_color, ansi_reset

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from multiprocessing.synchronize import Event


# Shared memory configuration
ARRAY_SIZE = 1000  # Adjust based on your needs

SERVER_PORT = 50001
DEV_SERVER_PORT = 54321


def run_vite(
    shutdown_event: "Event",
    debug: bool = False,
    dev_port: int = DEV_SERVER_PORT,
    backend_port: int = SERVER_PORT,
    dev_env_vars: dict = {},
) -> None:
    """Run Vite in a separate shell"""
    import subprocess

    signal.signal(signal.SIGINT, signal.SIG_IGN)

    logger = create_logger(
        "Vite",
        "magenta",
        logging.DEBUG if debug else logging.INFO,
    )

    os.environ["VITE_BACKEND_PORT"] = str(backend_port)

    for key, value in dev_env_vars.items():
        os.environ["VITE_" + key] = value

    def output_reader(shutdown_event: "Event", process: subprocess.Popen):
        logger.debug("Initialized output stream")
        color = ansi_color(color="magenta", bright=True, styles=["BOLD"])

        def read_stream(stream, stream_name):
            for line in iter(stream.readline, ""):
                if shutdown_event.is_set():
                    break
                logger.info(f"{color}[{stream_name}]{ansi_reset} {line.strip()}")

        stdout_thread = threading.Thread(
            target=read_stream, args=(process.stdout, "stdout")
        )
        stderr_thread = threading.Thread(
            target=read_stream, args=(process.stderr, "stderr")
        )

        stdout_thread.start()
        stderr_thread.start()

        shutdown_event.wait()

        stdout_thread.join(timeout=2)
        stderr_thread.join(timeout=2)

        logger.debug("Output stream closed")

    # Handle different operating systems
    shutdown_signal = signal.CTRL_BREAK_EVENT if os.name == "nt" else signal.SIGINT
    subprocess_flags = subprocess.CREATE_NEW_PROCESS_GROUP if os.name == "nt" else 0

    process = subprocess.Popen(
        "bun run dev --port " + str(dev_port),
        cwd="gui_dev",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        creationflags=subprocess_flags,
        shell=True,
    )

    logging_thread = threading.Thread(
        target=output_reader,
        args=(shutdown_event, process),
    )
    logging_thread.start()

    shutdown_event.wait()  # Wait for shutdown

    logger.debug("Terminating Vite server...")
    process.send_signal(shutdown_signal)

    try:
        process.wait(timeout=3)
    except subprocess.TimeoutExpired:
        logger.debug("Timeout expired, forcing termination...")
        process.kill()

    logging_thread.join(timeout=3)
    if logging_thread.is_alive():
        logger.debug("Logging thread did not finish in time")

    logger.info("Development server stopped")


def run_uvicorn(
    debug: bool = False, reload=False, server_port: int = SERVER_PORT
) -> None:
    from uvicorn.server import Server
    from uvicorn.config import LOGGING_CONFIG, Config

    # Configure logging
    color = ansi_color(color="green", bright=True, styles=["BOLD"])
    log_level = "DEBUG" if debug else "INFO"
    log_config = LOGGING_CONFIG.copy()
    log_config["loggers"]["uvicorn"]["level"] = log_level
    log_config["loggers"]["uvicorn.error"]["level"] = log_level
    log_config["loggers"]["uvicorn.access"]["level"] = log_level
    log_config["formatters"]["default"]["fmt"] = (
        f"{color}[FastAPI %(levelname)s (%(asctime)s)]:{ansi_reset} %(message)s"
    )
    log_config["formatters"]["default"]["datefmt"] = "%H:%M:%S"
    log_config["formatters"]["access"]["fmt"] = (
        f"{color}[FastAPI access (%(asctime)s)]:{ansi_reset} %(message)s"
    )
    log_config["formatters"]["access"]["datefmt"] = "%H:%M:%S"

    config = Config(
        app="py_neuromodulation.gui.backend.app_backend:PyNMBackend",
        host="localhost",
        reload=reload,
        factory=True,
        port=server_port,
        log_level="debug" if debug else "info",
        log_config=log_config,
        http="httptools",
        ws_ping_interval=None,
        ws_ping_timeout=None,
        ws_max_size=1024 * 1024 * 1024,  # 1GB
        loop="asyncio" if platform.system() == "Windows" else "uvloop",
        lifespan="off",
    )

    server = Server(config=config)

    if reload:
        from uvicorn.supervisors import ChangeReload
        from uvicorn._subprocess import get_subprocess

        # Overload the restart method of uvicorn so that is does not kill all of our processes
        # IMPORTANT: This is a hack and prevents shutdown events from triggering when the reloader is used
        class CustomReloader(ChangeReload):
            def restart(self) -> None:
                self.process.terminate()  # Use terminate instead of os.kill
                self.process.join()
                self.process = get_subprocess(
                    config=self.config, target=self.target, sockets=self.sockets
                )
                self.process.start()

        sock = config.bind_socket()
        server = CustomReloader(config, target=server.run, sockets=[sock])

    server.run()


def run_backend(
    shutdown_event: "Event",
    dev: bool = True,
    debug: bool = False,
    reload: bool = True,
    server_port: int = SERVER_PORT,
    dev_port: int = DEV_SERVER_PORT,
) -> None:
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    # Pass create_backend parameters through environment variables
    os.environ["PYNM_DEBUG"] = str(debug)
    os.environ["PYNM_DEV"] = str(dev)
    os.environ["PYNM_DEV_PORT"] = str(dev_port)

    server_process = mp.Process(
        target=run_uvicorn,
        kwargs={"debug": debug, "reload": reload, "server_port": server_port},
        name="Server",
    )
    server_process.start()
    shutdown_event.wait()
    server_process.join()


class AppManager:
    LAUNCH_FLAG = "PYNM_RUNNING"

    def __init__(
        self,
        debug: bool = False,
        dev: bool = True,
        run_in_webview=False,
        server_port=SERVER_PORT,
        dev_port=DEV_SERVER_PORT,
        dev_env_vars: dict = {},
    ) -> None:
        """_summary_

        Args:
            debug (bool, optional): If True, run the app in debug mode, which sets logging level to debug,
                and starts uvicorn, FastAPI, and Vite in debug mode. Defaults to False.
            dev (bool, optional): If True, run the app in development mode, which enables hot
                reloading and runs the frontend in Vite server. If False, run the app in production mode,
                which runs the frontend from the static files in the `frontend` directory. Defaults to True.
            run_in_webview (bool, optional): If True, open a PyWebView window to display the app. Defaults to False.
        """
        self.debug = debug
        self.dev = dev
        self.run_in_webview = run_in_webview
        self.server_port = server_port
        self.dev_port = dev_port
        self.dev_env_vars = dev_env_vars

        self._reset()
        # Prevent launching multiple instances of the app due to multiprocessing
        # This allows the absence of a main guard in the main script
        self.is_child_process = os.environ.get(self.LAUNCH_FLAG) == "TRUE"
        os.environ[self.LAUNCH_FLAG] = "TRUE"

        self.logger = create_logger(
            "PyNM",
            "yellow",
            logging.DEBUG if self.debug else logging.INFO,
        )

    def _reset(self) -> None:
        """Reset the AppManager to its initial state."""
        # Flags to track the state of the application
        self.shutdown_complete = False
        self.shutdown_started = False

        # Store background tasks
        self.tasks: dict[str, mp.Process] = {}

        # Events for multiprocessing synchronization
        self.shutdown_event: Event = mp.Event()

    def _terminate_app(self) -> None:
        if self.shutdown_started:
            self.logger.info("Termination already in progress. Skipping.")
            return

        self.shutdown_started = True

        timeout = 10
        deadline = time.time() + timeout

        self.logger.info("App closed, cleaning up background tasks...")
        self.shutdown_event.set()

        for process_name, process in self.tasks.items():
            remaining_time = max(deadline - time.time(), 0)
            process.join(timeout=remaining_time)

            if process.is_alive():
                self.logger.info(
                    f"{process_name} did not terminate in time. Forcing termination..."
                )
                force_terminate_process(process, process_name, logger=self.logger)

            self.logger.info(f"Process {process.name} terminated.")

        self.shutdown_complete = True
        self.shutdown_event.clear()
        self.logger.info("All background tasks succesfully terminated.")

    def _sigint_handler(self, signum, frame):
        if not self.shutdown_started:
            self.logger.info("Received SIGINT. Initiating graceful shutdown...")
            self._terminate_app()
        else:
            self.logger.info("SIGINT received again. Ignoring...")

    def launch(self) -> None:
        if self.is_child_process:
            return

        # Handle keyboard interrupt signals
        signal.signal(signal.SIGINT, self._sigint_handler)
        # signal.signal(signal.SIGINT, signal.SIG_IGN)

        # Create and start the subprocesses
        if self.dev:
            self.logger.info("Starting Vite server...")
            self.tasks["vite"] = mp.Process(
                target=run_vite,
                kwargs={
                    "shutdown_event": self.shutdown_event,
                    "debug": self.debug,
                    "dev_port": self.dev_port,
                    "backend_port": self.server_port,
                    "dev_env_vars": self.dev_env_vars,
                },
                name="Vite",
            )

        self.logger.info("Starting backend server...")
        self.tasks["backend"] = mp.Process(
            target=run_backend,
            kwargs={
                "shutdown_event": self.shutdown_event,
                "debug": self.debug,
                "reload": self.dev,
                "dev": self.dev,
                "server_port": self.server_port,
                "dev_port": self.dev_port,
            },
            name="Backend",
        )

        for process in self.tasks.values():
            process.start()

        if self.run_in_webview:
            from .app_window import WebViewWindow

            self.logger.info("Starting PyWebView window...")
            window = WebViewWindow(debug=self.debug)  # Must be called from main thread
            window.register_event_handler("closed", self._terminate_app)
            window.start()  # Start the window, this will block until the window is closed
        else:
            try:
                while not self.shutdown_complete:
                    time.sleep(0.1)
            except KeyboardInterrupt:
                pass  # The SIGINT handler will take care of termination

        if not self.shutdown_complete:
            self._terminate_app()

        self.logger.info("All processes cleaned up. Exiting...")



import numpy as np

import py_neuromodulation as nm


def get_example_stream(test_arr: np.ndarray) -> nm.Stream:
    settings = nm.NMSettings.get_default().enable_all_features()
    settings.features.nolds = False
    settings.features.mne_connectivity = False
    settings.features.coherence = False

    channels = nm.utils.get_default_channels_from_data(test_arr)

    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)
    return stream


def test_all_features_random_array():
    """This test runs's through all enabled features, and check's if they break"""
    np.random.seed(0)
    arr = np.random.random([2, 2000])
    stream = get_example_stream(arr)

    df = stream.run(
        arr, out_dir="./test_data", experiment_name="test_all_features_random_array"
    )

    assert df.shape[0] != 0  # terrible test


def test_all_features_zero_array():
    arr = np.zeros([2, 2000])

    stream = get_example_stream(arr)
    stream.settings.features.fooof = False  # Can't use fooof with 0s (log(0) undefined)

    df = stream.run(
        arr, out_dir="./test_data", experiment_name="test_all_features_zero_array"
    )

    assert df.shape[0] != 0  # terrible test


def test_all_features_NaN_array():
    arr = np.empty([2, 2000])
    arr[:] = np.nan

    stream = get_example_stream(arr)
    stream.settings.features.fooof = False  # Can't use fooof nan values

    df = stream.run(
        arr, out_dir="./test_data", experiment_name="test_all_features_NaN_array"
    )

    assert df.shape[0] != 0  # terrible test



import numpy as np
from py_neuromodulation.utils.types import NMBaseModel
from py_neuromodulation.utils.pydantic_extensions import NMField
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings
    import pandas as pd


class ProjectionSettings(NMBaseModel):
    max_dist_mm: float = NMField(default=20.0, gt=0.0, custom_metadata={"unit": "mm"})


class Projection:
    def __init__(
        self,
        settings: "NMSettings",
        grid_cortex: "pd.DataFrame",
        grid_subcortex: "pd.DataFrame",
        coords: dict,
        channels: "pd.DataFrame",
        plot_projection: bool = False,
    ) -> None:
        self.grid_cortex = grid_cortex
        self.grid_subcortex = grid_subcortex
        self.coords = coords
        self.channels = channels
        self.project_cortex = settings.postprocessing.project_cortex
        self.project_subcortex = settings.postprocessing.project_subcortex
        self.max_dist_cortex = settings.project_cortex_settings.max_dist_mm
        self.max_dist_subcortex = settings.project_subcortex_settings.max_dist_mm
        self.ecog_channels: list  # None case never handled, no need for default value
        self.lfp_channels: list  # None case never handled, no need for default value

        self.idx_chs_ecog: list = []  # feature series indexes for ecog channels
        self.names_chs_ecog: list = []  # feature series name of ecog features
        self.idx_chs_lfp: list = []  # feature series indexes for lfp channels
        self.names_chs_lfp: list = []  # feature series name of lfp features
        self.feature_names: list | None = None
        self.initialized: bool = False

        self.remove_not_used_ch_from_coords()  # remove beforehand non used channels from coords

        if len(self.coords["cortex_left"]["positions"]) == 0:
            self.sess_right = True
            self.ecog_strip = self.coords["cortex_right"]["positions"]
            self.ecog_strip_names = self.coords["cortex_right"]["ch_names"]
        elif len(self.coords["cortex_right"]["positions"]) == 0:
            self.sess_right = False
            self.ecog_strip = self.coords["cortex_left"]["positions"]
            self.ecog_strip_names = self.coords["cortex_left"]["ch_names"]

        if self.sess_right and len(self.coords["subcortex_right"]["positions"]) > 0:
            self.lfp_elec = self.coords["subcortex_right"]["positions"]
            self.lfp_elec_names = self.coords["subcortex_right"]["ch_names"]
        elif (
            self.sess_right is False
            and len(self.coords["subcortex_left"]["positions"]) > 0
        ):
            self.lfp_elec = self.coords["subcortex_left"]["positions"]
            self.lfp_elec_names = self.coords["subcortex_left"]["ch_names"]

        self._initialize_channels()

        (
            self.proj_matrix_cortex,
            self.proj_matrix_subcortex,
        ) = self.calc_projection_matrix()

        if self.project_cortex:
            self.active_cortex_gridpoints = np.nonzero(
                self.proj_matrix_cortex.sum(axis=1)
            )[0]
        if self.project_subcortex:
            self.active_subcortex_gridpoints = np.nonzero(
                self.proj_matrix_subcortex.sum(axis=1)
            )[0]

        if plot_projection:
            from py_neuromodulation.analysis.plots import NM_Plot

            nmplotter = NM_Plot(
                ecog_strip=self.ecog_strip,
                grid_cortex=self.grid_cortex.to_numpy(),
                grid_subcortex=self.grid_subcortex.to_numpy(),
                sess_right=self.sess_right,
                proj_matrix_cortex=self.proj_matrix_cortex,
            )
            nmplotter.plot_cortex()

    def remove_not_used_ch_from_coords(self):
        ch_not_used = self.channels.query('(used==0) or (status=="bad")').name
        if len(ch_not_used) > 0:
            for ch in ch_not_used:
                for key_ in self.coords:
                    for idx, ch_coords in enumerate(self.coords[key_]["ch_names"]):
                        if ch.startswith(ch_coords):
                            # delete index
                            self.coords[key_]["positions"] = np.delete(
                                self.coords[key_]["positions"], idx, axis=0
                            )
                            self.coords[key_]["ch_names"].remove(ch)

    def calc_proj_matrix(
        self,
        max_dist: float,
        grid: np.ndarray,
        coord_array: np.ndarray,
    ) -> np.ndarray:
        """Calculate projection matrix."""

        channels = coord_array.shape[0]
        distance_matrix = np.zeros([grid.shape[1], channels])

        for project_point in range(grid.shape[1]):
            for channel in range(coord_array.shape[0]):
                distance_matrix[project_point, channel] = np.linalg.norm(
                    grid[:, project_point] - coord_array[channel, :]
                )

        proj_matrix = np.zeros(distance_matrix.shape)
        for grid_point in range(distance_matrix.shape[0]):
            used_channels = np.where(distance_matrix[grid_point, :] < max_dist)[0]

            rec_distances = distance_matrix[grid_point, used_channels]
            sum_distances: float = np.sum(1 / rec_distances)

            for _, used_channel in enumerate(used_channels):
                proj_matrix[grid_point, used_channel] = (
                    1 / distance_matrix[grid_point, used_channel]
                ) / sum_distances
        return proj_matrix

    def calc_projection_matrix(self):
        """Calculates a projection matrix based on the used coordinate arrays

        Returns
        -------
        proj_matrix_cortex (np.array)
            cortical projection_matrix in shape [grid contacts, channel contact] defaults to None
        proj_matrix_subcortex (np.array)
            subcortical projection_matrix in shape [grid contacts, channel contact] defaults to None
        """

        proj_matrix_run = np.empty(2, dtype=object)

        if self.sess_right:
            if self.project_cortex:
                cortex_grid_right = np.copy(self.grid_cortex)
                cortex_grid_right[:, 0] = cortex_grid_right[:, 0] * -1
                self.cortex_grid_right = np.array(cortex_grid_right.T)
            else:
                self.cortex_grid_right = None

            if self.project_subcortex:
                subcortex_grid_right = np.copy(self.grid_subcortex)
                subcortex_grid_right[:, 0] = subcortex_grid_right[:, 0] * -1
                self.subcortex_grid_right = np.array(subcortex_grid_right).T
            else:
                self.subcortex_grid_right = None

            grid_session = [self.cortex_grid_right, self.subcortex_grid_right]

        else:
            if self.project_cortex:
                self.cortex_grid_left = np.array(self.grid_cortex.T)
            else:
                self.cortex_grid_left = None
            if self.project_subcortex:
                self.subcortex_grid_left = np.array(self.grid_subcortex.T)
            else:
                self.subcortex_grid_left = None

            grid_session = [self.cortex_grid_left, self.subcortex_grid_left]

        coord_array = [
            self.ecog_strip if grid_session[0] is not None else None,
            self.lfp_elec if grid_session[1] is not None else None,
        ]

        for loc_, grid in enumerate(grid_session):
            if loc_ == 0:  # cortex
                max_dist = self.max_dist_cortex
            elif loc_ == 1:  # subcortex
                max_dist = self.max_dist_subcortex

            if grid_session[loc_] is not None:
                proj_matrix_run[loc_] = self.calc_proj_matrix(
                    max_dist, grid, coord_array[loc_]
                )

        return proj_matrix_run[0], proj_matrix_run[1]  # cortex, subcortex

    def _initialize_channels(self) -> None:
        """Initialize channel names via nm_channel new_name column"""

        if self.project_cortex:
            self.ecog_channels = self.channels.query(
                '(type=="ecog") and (used == 1) and (status=="good")'
            ).name.to_list()

            chs_ecog = self.ecog_channels.copy()
            for ecog_channel in chs_ecog:
                if ecog_channel not in self.ecog_strip_names:
                    self.ecog_channels.remove(ecog_channel)
            # write ecog_channels to be new_name
            self.ecog_channels = list(
                self.channels.query("name == @self.ecog_channels").new_name
            )

        if self.project_subcortex:
            self.lfp_channels = self.channels.query(
                '(type=="lfp" or type=="seeg" or type=="dbs") \
              and (used == 1) and (status=="good")'
            ).name.to_list()
            # project only channels that are in the coords
            # this also deletes channels of the other hemisphere
            chs_lfp = self.lfp_channels.copy()
            for lfp_channel in chs_lfp:
                if lfp_channel not in self.lfp_elec_names:
                    self.lfp_channels.remove(lfp_channel)
            # write lfp_channels to be new_name
            self.lfp_channels = list(
                self.channels.query("name == @self.lfp_channels").new_name
            )

    def init_projection_run(self, feature_dict: dict) -> None:
        """Initialize indexes for respective channels in feature series computed by features.py"""
        #  here it is assumed that only one hemisphere is recorded at a time!
        if self.project_cortex:
            for ecog_channel in self.ecog_channels:
                self.idx_chs_ecog.append(
                    [
                        ch_idx
                        for ch_idx, ch in enumerate(feature_dict.keys())
                        if ch.startswith(ecog_channel)
                    ]
                )
                self.names_chs_ecog.append(
                    [
                        ch
                        for _, ch in enumerate(feature_dict.keys())
                        if ch.startswith(ecog_channel)
                    ]
                )
            if self.names_chs_ecog:
                # get feature_names; given by ECoG sequency of features
                self.feature_names = [
                    feature_name[len(self.ecog_channels[0]) + 1 :]
                    for feature_name in self.names_chs_ecog[0]
                ]

        if self.project_subcortex:
            # for lfp_channels select here only the ones from the correct hemisphere!
            for lfp_channel in self.lfp_channels:
                self.idx_chs_lfp.append(
                    [
                        ch_idx
                        for ch_idx, ch in enumerate(feature_dict.keys())
                        if ch.startswith(lfp_channel)
                    ]
                )
                self.names_chs_lfp.append(
                    [
                        ch
                        for _, ch in enumerate(feature_dict.keys())
                        if ch.startswith(lfp_channel)
                    ]
                )
            if not self.feature_names and self.names_chs_lfp:
                # get feature_names; given by LFP sequency of features
                self.feature_names = [
                    feature_name[len(self.lfp_channels[0]) + 1 :]
                    for feature_name in self.names_chs_lfp[0]
                ]

        self.initialized = True

    def project_features(self, feature_dict: dict) -> None:
        """Project data, given idx_chs_ecog/stn"""

        if not self.initialized:
            self.init_projection_run(feature_dict)

        dat_cortex = (
            np.array(
                [
                    np.array([feature_dict[ch] for ch in ch_names])
                    for ch_names in self.names_chs_ecog
                ]
            )
            if self.project_cortex
            else None
        )

        dat_subcortex = (
            np.array(
                [
                    np.array([feature_dict[ch] for ch in ch_names])
                    for ch_names in self.names_chs_lfp
                ]
            )
            if self.project_subcortex
            else None
        )

        # project data
        # get_projected_cortex_subcortex_data can return None
        # but None is not handled and will throw error in the code below

        proj_cortex_array: np.ndarray
        proj_subcortex_array: np.ndarray
        (proj_cortex_array, proj_subcortex_array) = (
            self.get_projected_cortex_subcortex_data(dat_cortex, dat_subcortex)
        )  # type: ignore # Ignore None return

        features_new: dict = {}
        # proj_cortex_array has shape grid_points x feature_number
        if self.project_cortex:
            features_new = features_new | {
                "gridcortex_"
                + str(act_grid_point)
                + "_"
                + feature_name: proj_cortex_array[act_grid_point, feature_idx]
                for feature_idx, feature_name in enumerate(self.feature_names)  # type: ignore # Empty list handled above
                for act_grid_point in self.active_cortex_gridpoints
            }
        if self.project_subcortex:
            features_new = features_new | {
                "gridsubcortex_"
                + str(act_grid_point)
                + "_"
                + feature_name: proj_subcortex_array[act_grid_point, feature_idx]
                for feature_idx, feature_name in enumerate(self.feature_names)  # type: ignore # Empty list handled above
                for act_grid_point in self.active_subcortex_gridpoints
            }

        feature_dict.update(features_new)

    def get_projected_cortex_subcortex_data(
        self,
        dat_cortex: np.ndarray | None = None,
        dat_subcortex: np.ndarray | None = None,
    ) -> tuple[np.ndarray | None, np.ndarray | None]:
        """Project cortical and subcortical data to predefined projection matrices

        Parameters
        ----------
        dat_cortex : np.ndarray, optional
            cortical features, by default None
        dat_subcortex : np.ndarray, optional
            subcortical features, by default None

        Returns
        -------
        proj_cortex : np.ndarray
            projected cortical features, by detault None
        proj_subcortex : np.ndarray
            projected subcortical features, by detault None
        """
        proj_cortex = None
        proj_subcortex = None

        if dat_cortex is not None:
            proj_cortex = self.proj_matrix_cortex @ dat_cortex
        if dat_subcortex is not None:
            proj_subcortex = self.proj_matrix_subcortex @ dat_subcortex

        return proj_cortex, proj_subcortex



import numpy as np

from typing import TYPE_CHECKING

from py_neuromodulation.utils.types import BoolSelector, FrequencyRange, NMPreprocessor
from py_neuromodulation.utils.pydantic_extensions import NMField

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings


class FilterSettings(BoolSelector):
    bandstop_filter: bool = True
    bandpass_filter: bool = True
    lowpass_filter: bool = True
    highpass_filter: bool = True

    bandstop_filter_settings: FrequencyRange = FrequencyRange(100, 160)
    bandpass_filter_settings: FrequencyRange = FrequencyRange(2, 200)
    lowpass_filter_cutoff_hz: float = NMField(
        default=200, gt=0, custom_metadata={"unit": "Hz"}
    )
    highpass_filter_cutoff_hz: float = NMField(
        default=3, gt=0, custom_metadata={"unit": "Hz"}
    )

    def get_filter_tuple(self, filter_name) -> FrequencyRange:
        match filter_name:
            case "bandstop_filter":
                return self.bandstop_filter_settings
            case "bandpass_filter":
                return self.bandpass_filter_settings
            case "lowpass_filter":
                return FrequencyRange(None, self.lowpass_filter_cutoff_hz)
            case "highpass_filter":
                return FrequencyRange(self.highpass_filter_cutoff_hz, None)
            case _:
                raise ValueError(
                    "Filter name must be one of 'bandstop_filter', 'lowpass_filter', "
                    "'highpass_filter', 'bandpass_filter'"
                )


class PreprocessingFilter(NMPreprocessor):
    def __init__(self, settings: "NMSettings", sfreq: float) -> None:
        from py_neuromodulation.filter import MNEFilter


        self.filters: list[MNEFilter] = []
        for filter_name in settings.preprocessing_filter.get_enabled():
            if filter_name != "lowpass_filter" and filter_name != "highpass_filter":
                self.filters += [
                    MNEFilter(
                        f_ranges=[settings.preprocessing_filter.get_filter_tuple(filter_name)],  # type: ignore
                        sfreq=sfreq,
                        filter_length=sfreq - 1,
                        verbose=False,
                    )
                ]

        if "lowpass_filter" in settings.preprocessing_filter.get_enabled():
            self.filters.append(
                MNEFilter(
                    f_ranges=[[None, settings.preprocessing_filter.lowpass_filter_cutoff_hz]],  # type: ignore
                    sfreq=sfreq,
                    filter_length=sfreq - 1,
                    verbose=False,
                )
            )
        if "highpass_filter" in settings.preprocessing_filter.get_enabled():
            self.filters.append(
                MNEFilter(
                    f_ranges=[[settings.preprocessing_filter.highpass_filter_cutoff_hz, None]],  # type: ignore
                    sfreq=sfreq,
                    filter_length=sfreq - 1,
                    verbose=False,
                )
            )

    def process(self, data: np.ndarray) -> np.ndarray:
        """Preprocess data according to the initialized list of PreprocessingFilter objects

        Args:
            data (numpy ndarray) :
                shape(n_channels, n_samples) - data to be preprocessed.

        Returns:
            preprocessed_data (numpy ndarray):
            shape(n_channels, n_samples) - preprocessed data
        """

        for filter in self.filters:
            data = filter.filter_data(data if len(data.shape) == 2 else data[:, 0, :])
        return data if len(data.shape) == 2 else data[:, 0, :]



"""
Real-time feature estimation
============================

"""

# %%
# Implementation of individual nm_streams
# ---------------------------------------
#
# *py_neuromodulation* was optimized for computation of real-time data streams.
# There are however center -and lab specific hardware acquisition systems. Therefore, each experiment requires modules to interact with hardware platforms
# which periodically acquire data.
#
# Given the raw data, data can be analyzed using *py_neuromodulation*. Preprocessing methods, such as re-referencing and normalization,
# feature computation and decoding can be performed then in real-time.
#
# For online as well as as offline analysis, the :class:`~nm_stream_abc` class needs to be instantiated.
# Here the `nm_settings` and `channels` are required to be defined.
# Previously for the offline analysis, an offline :class:`~nm_generator` object was defined that periodically yielded data.
# For online data, the :meth:`~stream_abc.run` function therefore needs to be overwritten, which first acquires data and then calls
# the :meth:`~run_analysis.process` function.
#
# The following illustrates in pseudo-code how such a stream could be initialized:
#
# .. code-block:: python
#
#    from py_neuromodulation import nm_stream_abc
#
#    class MyStream(nm_stream_abc):
#    def __init__(self, settings, channels):
#        super().__init__(settings, channels)
#
#    def run(self):
#       features_ = []
#        while True:
#           data = self.acquire_data()
#           features_.append(self.run_analysis.process(data))
#           # potentially use machine learning model for decoding
#
#
# Computation time examples
# -------------------------
#
# The following example calculates for six channels, CAR re-referencing, z-score normalization and FFT features results the following computation time:

# %%
import py_neuromodulation as nm
from py_neuromodulation import NMSettings
import numpy as np
import timeit


def get_fast_compute_settings():
    settings = NMSettings.get_fast_compute()

    settings.preprocessing = ["re_referencing", "notch_filter"]
    settings.features.fft = True
    settings.postprocessing.feature_normalization = True
    return settings


data = np.random.random([1, 1000])

print("FFT Features, CAR re-referencing, z-score normalization")
print()
print("Computation time for single ECoG channel: ")
stream = nm.Stream(
    sfreq=1000,
    data=data,
    sampling_rate_features_hz=10,
    verbose=False,
    settings=get_fast_compute_settings(),
)
print(
    f"{np.round(timeit.timeit(lambda: stream.data_processor.process(data), number=100)/100, 3)} s"
)

print("Computation time for 6 ECoG channels: ")
data = np.random.random([6, 1000])
stream = nm.Stream(
    sfreq=1000,
    data=data,
    sampling_rate_features_hz=10,
    verbose=False,
    settings=get_fast_compute_settings(),
)
print(
    f"{np.round(timeit.timeit(lambda: stream.data_processor.process(data), number=100)/100, 3)} s"
)

print(
    "\nFFT Features & Temporal Waveform Shape & Hjorth & Bursts, CAR re-referencing, z-score normalization"
)
print("Computation time for single ECoG channel: ")
data = np.random.random([1, 1000])
stream = nm.Stream(sfreq=1000, data=data, sampling_rate_features_hz=10, verbose=False)
print(
    f"{np.round(timeit.timeit(lambda: stream.data_processor.process(data), number=10)/10, 3)} s"
)


# %%
# Those results show that the computation time for a typical pipeline (FFT, re-referencing, notch-filtering, feature normalization)
# is well below 10 ms, which is fast enough for real-time analysis with feature sampling rates below 100 Hz.
# Computation of more complex features could still result in feature sampling rates of more than 30 Hz.
#
# Real-time movement decoding using the TMSi-SAGA amplifier
# ---------------------------------------------------------
#
# In the following example, we will show how we setup a real-time movement decoding experiment using the TMSi-SAGA amplifier.
# First, we relied on different software modules for data streaming and visualization.
# `LabStreamingLayer <https://labstreaminglayer.org>`_ allows for real-time data streaming and synchronization across multiple devices.
# We used `timeflux <https://timeflux.io>`_ for real-time data visualization of features, decoded output.
# For raw data visualization we used `Brain Streaming Layer <https://fcbg-hnp-meeg.github.io/bsl/dev/index.html>`_.
#
# The code for real-time movement decoding is added in the GitHub branch `realtime_decoding <https://github.com/neuromodulation/py_neuromodulation/tree/realtime_decoding>`_.
# Here we relied on the `TMSI SAGA Python interface <https://gitlab.com/tmsi/tmsi-python-interface>`_.
#



import numpy as np


class RawDataGenerator:
    """
    This generator function mimics online data acquisition.
    The data are iteratively sampled with settings.sampling_rate_features_hz
    """

    def __init__(
        self,
        data: np.ndarray,
        sfreq: float,
        sampling_rate_features_hz: float,
        segment_length_features_ms: float,
    ) -> None:
        """
        Arguments
        ---------
            data (np array): shape (channels, time)
            settings (settings.NMSettings): settings object
            sfreq (float): sampling frequency of the data

        Returns
        -------
            np.array: 1D array of time stamps
            np.array: new batch for run function of full segment length shape
        """
        self.batch_counter: int = 0  # counter for the batches

        self.data = data
        self.sfreq = sfreq
        # Width, in data points, of the moving window used to calculate features
        self.segment_length = segment_length_features_ms / 1000 * sfreq
        # Ratio of the sampling frequency of the input data to the sampling frequency
        self.stride = sfreq / sampling_rate_features_hz

    def __iter__(self):
        return self

    def __next__(self):
        start = self.stride * self.batch_counter
        end = start + self.segment_length

        self.batch_counter += 1

        start_idx = int(start)
        end_idx = int(end)

        if end_idx > self.data.shape[1]:
            raise StopIteration

        return np.arange(start, end) / self.sfreq, self.data[:, start_idx:end_idx]



from os import PathLike
from math import isnan
from typing import Literal, TYPE_CHECKING, Any
from pydantic import BaseModel, model_validator, Field
from .pydantic_extensions import NMBaseModel, NMField
from abc import abstractmethod

from collections.abc import Sequence
from datetime import datetime


if TYPE_CHECKING:
    import numpy as np
    from py_neuromodulation import NMSettings

###################################
########## TYPE ALIASES  ##########
###################################

_PathLike = str | PathLike

FEATURE_NAME = Literal[
    "raw_hjorth",
    "return_raw",
    "bandpass_filter",
    "stft",
    "fft",
    "welch",
    "sharpwave_analysis",
    "fooof",
    "nolds",
    "coherence",
    "bursts",
    "linelength",
    "mne_connectivity",
    "bispectrum",
]

PREPROCESSOR_NAME = Literal[
    "preprocessing_filter",
    "notch_filter",
    "raw_resampling",
    "re_referencing",
    "raw_normalization",
]

NORM_METHOD = Literal[
    "mean",
    "median",
    "zscore",
    "zscore-median",
    "quantile",
    "power",
    "robust",
    "minmax",
]


class NMFeature:
    def __init__(
        self, settings: "NMSettings", ch_names: Sequence[str], sfreq: int | float
    ) -> None: ...

    def calc_feature(self, data: "np.ndarray") -> dict:
        """
        Feature calculation method. Each method needs to loop through all channels

        Parameters
        ----------
        data : 'np.ndarray'
            (channels, time)

        Returns
        -------
        dict
        """
        ...


class NMPreprocessor:
    def process(self, data: "np.ndarray") -> "np.ndarray": ...


class FrequencyRange(NMBaseModel):
    frequency_low_hz: float = Field(gt=0)
    frequency_high_hz: float = Field(gt=0)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def __getitem__(self, item: int):
        match item:
            case 0:
                return self.frequency_low_hz
            case 1:
                return self.frequency_high_hz
            case _:
                raise IndexError(f"Index {item} out of range")

    def as_tuple(self) -> tuple[float, float]:
        return (self.frequency_low_hz, self.frequency_high_hz)

    def __iter__(self):  # type: ignore
        return iter(self.as_tuple())

    @model_validator(mode="after")
    def validate_range(self):
        if not (isnan(self.frequency_high_hz) or isnan(self.frequency_low_hz)):
            assert self.frequency_high_hz > self.frequency_low_hz, (
                "Frequency high must be greater than frequency low"
            )
        return self

    @model_validator(mode="before")
    @classmethod
    def check_input(cls, input):
        """Pydantic validator to convert the input to a dictionary when passed as a list
        as we have it by default in the default_settings.yaml file
        For example, [1,2] will be converted to {"frequency_low_hz": 1, "frequency_high_hz": 2}
        """
        match input:
            case dict() if "frequency_low_hz" in input and "frequency_high_hz" in input:
                return input
            case Sequence() if len(input) == 2:
                return {"frequency_low_hz": input[0], "frequency_high_hz": input[1]}
            case _:
                raise ValueError(
                    "Value for FrequencyRange must be a dictionary, "
                    "or a sequence of 2 numeric values, "
                    f"but got {input} instead."
                )


class BoolSelector(NMBaseModel):
    def get_enabled(self):
        return [
            f
            for f in self.model_fields.keys()
            if (isinstance(self[f], bool) and self[f])
        ]

    def enable_all(self):
        for f in self.model_fields.keys():
            if isinstance(self[f], bool):
                self[f] = True

    def disable_all(self):
        for f in self.model_fields.keys():
            if isinstance(self[f], bool):
                self[f] = False

    def __iter__(self):  # type: ignore
        return iter(self.model_dump().keys())

    @classmethod
    def list_all(cls):
        return list(cls.model_fields.keys())

    @classmethod
    def print_all(cls):
        for f in cls.list_all():
            print(f)


#################
### GUI TYPES ###
#################


class FileInfo(BaseModel):
    name: str
    path: str
    dir: str
    is_directory: bool
    size: int
    created_at: datetime
    modified_at: datetime


from pydantic import ValidationError
import pytest
import numpy as np

from py_neuromodulation import NMSettings, Stream, features
from py_neuromodulation.utils.types import FEATURE_NAME


def setup_osc_settings(
    osc_feature_name: FEATURE_NAME,
    osc_feature_setting: str,
    windowlength_ms: int,
    log_transform: bool,
):
    settings = NMSettings.get_default().reset()

    settings.features[osc_feature_name] = True
    settings[osc_feature_setting].windowlength_ms = windowlength_ms
    settings[osc_feature_setting].log_transform = log_transform

    return settings


def setup_bandpass_settings(log_transform: bool):
    settings = NMSettings.get_default().reset()

    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.log_transform = log_transform

    return settings


def test_fft_wrong_logtransform_param_init():
    """Test that settings cannot be initialized with wrong log_transform parameter value"""
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    with pytest.raises(ValidationError):
        settings = setup_osc_settings(
            osc_feature_name="fft",
            osc_feature_setting="fft_settings",
            windowlength_ms=1000,
            log_transform="123",
        )
        settings.frequency_ranges_hz = {"theta": (4, 8), "beta": (10, 20)}
        features.FFT(settings, ch_names, sfreq)


def test_fft_frequencyband_range_passing_nyquist_range():
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000
    data = np.random.random([len(ch_names), sfreq])

    settings = setup_osc_settings(
        osc_feature_name="fft",
        osc_feature_setting="fft_settings",
        windowlength_ms=1000,
        log_transform=False,
    )

    settings.frequency_ranges_hz = {"theta": [4, 8], "broadband": [10, 600]}

    with pytest.raises(AssertionError):
        Stream(sfreq=sfreq, data=data, settings=settings)


def test_fft_zero_data():
    """ """
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = setup_osc_settings(
        osc_feature_name="fft",
        osc_feature_setting="fft_settings",
        windowlength_ms=1000,
        log_transform=False,
    )
    settings.frequency_ranges_hz = {"theta": [4, 8], "beta": [10, 20]}
    fft_obj = features.FFT(settings, ch_names, sfreq)

    data = np.ones([len(ch_names), sfreq])
    features_out = fft_obj.calc_feature(data)

    for f in features_out.keys():
        if "psd_0" not in f:
            assert np.isclose(features_out[f], 0, atol=1e-6)


def test_fft_random_data():
    """Test that FFT feature extraction works with random numbers"""
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = setup_osc_settings(
        osc_feature_name="fft",
        osc_feature_setting="fft_settings",
        windowlength_ms=1000,
        log_transform=False,
    )
    settings.frequency_ranges_hz = {"theta": [4, 8], "beta": [10, 20]}
    fft_obj = features.FFT(settings, ch_names, sfreq)

    data = np.random.random([len(ch_names), sfreq])
    features_out = fft_obj.calc_feature(data)

    for f in features_out.keys():
        assert features_out[f] != 0


def test_fft_beta_osc():
    ch_names = [
        "ch1",
    ]
    sfreq = 1000

    settings = setup_osc_settings(
        osc_feature_name="fft",
        osc_feature_setting="fft_settings",
        windowlength_ms=1000,
        log_transform=False,
    )

    settings.frequency_ranges_hz = {
        "theta": [4, 8],
        "beta": [10, 28],
        "gamma": [50, 60],
    }

    fft_obj = features.FFT(settings, ch_names, sfreq)

    time_duration = 1

    time_points = np.arange(0, time_duration, 1 / sfreq)
    beta_freq = 20

    beta_wave = np.sin(2 * np.pi * beta_freq * time_points)

    np.random.seed(0)
    data = np.random.random([len(ch_names), sfreq]) + beta_wave

    features_out = fft_obj.calc_feature(data)

    assert (
        features_out["ch1_fft_beta_mean"] > features_out["ch1_fft_theta_mean"]
        and features_out["ch1_fft_beta_mean"] > features_out["ch1_fft_gamma_mean"]
    )


def test_stft_wrong_logtransform_param_init():
    """ """
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    with pytest.raises(ValidationError):
        settings = setup_osc_settings(
            osc_feature_name="stft",
            osc_feature_setting="stft_settings",
            windowlength_ms=1000,
            log_transform="123",
        )
        features.STFT(settings, ch_names, sfreq)


def test_stft_wrong_frequencyband_range_init():
    """ """
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000
    data = np.random.random([len(ch_names), sfreq])

    settings = setup_osc_settings(
        osc_feature_name="stft",
        osc_feature_setting="stft_settings",
        windowlength_ms=1000,
        log_transform=False,
    )
    settings.frequency_ranges_hz = {"theta": [4, 8], "broadband": [10, 600]}

    with pytest.raises(AssertionError):
        Stream(settings=settings, data=data, sfreq=sfreq)


def test_stft_beta_osc():
    ch_names = [
        "ch1",
    ]
    sfreq = 1000

    settings = setup_osc_settings(
        osc_feature_name="stft",
        osc_feature_setting="stft_settings",
        windowlength_ms=1000,
        log_transform=True,
    )

    settings.frequency_ranges_hz = {
        "theta": [4, 8],
        "beta": [10, 28],
        "gamma": [50, 60],
    }

    stft_obj = features.STFT(settings, ch_names, sfreq)

    time_duration = 1

    time_points = np.arange(0, time_duration, 1 / sfreq)
    beta_freq = 20

    beta_wave = np.sin(2 * np.pi * beta_freq * time_points)

    np.random.seed(0)
    data = np.random.random([len(ch_names), sfreq]) + beta_wave

    features_out = stft_obj.calc_feature(data)

    assert (
        features_out["ch1_stft_beta_mean"] > features_out["ch1_stft_theta_mean"]
        and features_out["ch1_stft_beta_mean"] > features_out["ch1_stft_gamma_mean"]
    )


def test_welch_beta_osc():
    ch_names = [
        "ch1",
    ]
    sfreq = 1000

    settings = setup_osc_settings(
        osc_feature_name="welch",
        osc_feature_setting="welch_settings",
        windowlength_ms=1000,
        log_transform=True,
    )

    settings.frequency_ranges_hz = {
        "theta": [4, 8],
        "beta": [10, 28],
        "gamma": [50, 60],
    }

    stft_obj = features.Welch(settings, ch_names, sfreq)

    time_duration = 1

    time_points = np.arange(0, time_duration, 1 / sfreq)
    beta_freq = 20

    beta_wave = np.sin(2 * np.pi * beta_freq * time_points)

    np.random.seed(0)
    data = np.random.random([len(ch_names), sfreq]) + beta_wave

    features_out = stft_obj.calc_feature(data)

    assert (
        features_out["ch1_welch_beta_mean"] > features_out["ch1_welch_theta_mean"]
        and features_out["ch1_welch_beta_mean"] > features_out["ch1_welch_gamma_mean"]
    )


def test_bp_wrong_logtransform_param_init():
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = NMSettings.get_default().reset()
    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.log_transform = "123"

    settings.frequency_ranges_hz = {"theta": [4, 8], "beta": [10, 20]}

    with pytest.raises(ValidationError):
        features.BandPower(settings, ch_names, sfreq)


def test_bp_wrong_frequencyband_range_init():
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = NMSettings.get_default().reset()
    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.log_transform = False

    settings.frequency_ranges_hz = {"theta": [4, 8], "broadband": [10, 600]}

    with pytest.raises(ValidationError):
        features.BandPower(settings, ch_names, sfreq)


def test_bp_non_defined_fband():
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = NMSettings.get_default().reset()
    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.log_transform = False
    settings.frequency_ranges_hz = {"theta": [4, 8], "broadband": [10, 600]}

    settings.bandpass_filter_settings.segment_lengths_ms["theta"] = 1000
    settings.bandpass_filter_settings.segment_lengths_ms["beta"] = 300

    with pytest.raises(ValidationError):
        features.BandPower(settings, ch_names, sfreq)


def test_bp_segment_length_fb_exceeds_segment_length_features():
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = NMSettings.get_default().reset()
    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.log_transform = False

    settings.segment_length_features_ms = 500
    settings.frequency_ranges_hz = {"theta": [4, 8], "broadband": [10, 600]}
    settings.bandpass_filter_settings.segment_lengths_ms["theta"] = 1000
    settings.bandpass_filter_settings.segment_lengths_ms["beta"] = 300

    with pytest.raises(Exception):
        features.BandPower(settings, ch_names, sfreq)


def test_bp_zero_data():
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = NMSettings.get_default().reset()
    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.segment_lengths_ms["theta"] = 1000
    settings.bandpass_filter_settings.segment_lengths_ms["beta"] = 300

    settings.bandpass_filter_settings.log_transform = False
    settings.bandpass_filter_settings.kalman_filter = False
    settings.bandpass_filter_settings.bandpower_features.activity = True

    settings.frequency_ranges_hz = {"theta": [4, 8], "beta": [10, 20]}
    stft_obj = features.BandPower(settings, ch_names, sfreq)

    data = np.zeros([len(ch_names), sfreq])
    features_out = stft_obj.calc_feature(data)

    for f in features_out.keys():
        assert pytest.approx(0, 0.01) == features_out[f]


def test_bp_random_data():
    ch_names = ["ch1", "ch2", "ch3", "ch4"]
    sfreq = 1000

    settings = NMSettings.get_default().reset()

    settings.frequency_ranges_hz = {"theta": [4, 8], "beta": [10, 30]}
    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.segment_lengths_ms["theta"] = 1000
    settings.bandpass_filter_settings.segment_lengths_ms["beta"] = 300

    settings.bandpass_filter_settings.log_transform = False
    settings.bandpass_filter_settings.kalman_filter = False
    settings.bandpass_filter_settings.bandpower_features.activity = True

    stft_obj = features.BandPower(settings, ch_names, sfreq)

    np.random.seed(0)
    data = np.random.random([len(ch_names), sfreq])
    features_out = stft_obj.calc_feature(data)

    for f in features_out.keys():
        assert pytest.approx(0, 0.01) != features_out[f]


def test_bp_beta_osc():
    ch_names = [
        "ch1",
    ]
    sfreq = 1000

    settings = NMSettings.get_default().reset()

    settings.frequency_ranges_hz = {
        "theta": [4, 8],
        "beta": [10, 30],
        "gamma": [50, 60],
    }

    settings.features.bandpass_filter = True
    settings.bandpass_filter_settings.segment_lengths_ms["theta"] = 1000
    settings.bandpass_filter_settings.segment_lengths_ms["beta"] = 300
    settings.bandpass_filter_settings.segment_lengths_ms["gamma"] = 100

    settings.bandpass_filter_settings.log_transform = False
    settings.bandpass_filter_settings.kalman_filter = False
    settings.bandpass_filter_settings.bandpower_features.activity = True

    bp_obj = features.BandPower(settings, ch_names, sfreq)

    time_duration = 1

    time_points = np.arange(0, time_duration, 1 / sfreq)
    beta_freq = 20

    beta_wave = np.sin(2 * np.pi * beta_freq * time_points)

    np.random.seed(0)
    data = np.random.random([len(ch_names), sfreq]) + beta_wave

    features_out = bp_obj.calc_feature(data)

    assert (
        features_out["ch1_bandpass_activity_beta"]
        > features_out["ch1_bandpass_activity_theta"]
        and features_out["ch1_bandpass_activity_beta"]
        > features_out["ch1_bandpass_activity_gamma"]
    )



"""
Analyzing temporal features
===========================

"""

# %%
# Time series data can be characterized using oscillatory components, but assumptions of sinusoidality are for real data rarely fulfilled.
# See *"Brain Oscillations and the Importance of Waveform Shape"* `Cole et al 2017 <https://doi.org/10.1016/j.tics.2016.12.008>`_ for a great motivation.
# We implemented here temporal characteristics based on individual trough and peak relations,
# based on the :meth:~`scipy.signal.find_peaks` method. The function parameter *distance* can be specified in the *settings.yaml*.
# Temporal features can be calculated twice for troughs and peaks. In the settings, this can be specified by setting *estimate* to true
# in *detect_troughs* and/or *detect_peaks*. A statistical measure (e.g. mean, max, median, var) can be defined as a resulting feature from the peak and
# trough estimates using the *apply_estimator_between_peaks_and_troughs* setting.
#
# In py_neuromodulation the following characteristics are implemented:
#
# .. note::
#     The nomenclature is written here for sharpwave troughs, but detection of peak characteristics can be computed in the same way.
#
# -  prominence:
#    :math:`V_{prominence} = |\frac{V_{peak-left} + V_{peak-right}}{2}| - V_{trough}`
# -  sharpness:
#    :math:`V_{sharpnesss} = \frac{(V_{trough} - V_{trough-5 ms}) + (V_{trough} - V_{trough+5 ms})}{2}`
# -  rise and decay rise time
# -  rise and decay steepness
# -  width (between left and right peaks)
# -  interval (between troughs)
#
# Additionally, different filter ranges can be parametrized using the *filter_ranges_hz* setting.
# Filtering is necessary to remove high frequent signal fluctuations, but limits also the true estimation of sharpness and prominence due to signal smoothing.

from typing import cast
import seaborn as sb
from matplotlib import pyplot as plt
from scipy import signal
from scipy.signal import fftconvolve
import numpy as np

import py_neuromodulation as nm
from py_neuromodulation import NMSettings
from py_neuromodulation.features import SharpwaveAnalyzer


# %%
# We will first read the example ECoG data and plot the identified features on the filtered time series.

RUN_NAME, PATH_RUN, PATH_BIDS, PATH_OUT, datatype = nm.io.get_paths_example_data()

(
    raw,
    data,
    sfreq,
    line_noise,
    coord_list,
    coord_names,
) = nm.io.read_BIDS_data(PATH_RUN=PATH_RUN)

# %%
settings = NMSettings.get_fast_compute()

settings.features.fft = True
settings.features.bursts = False
settings.features.sharpwave_analysis = True
settings.features.coherence = False

settings.sharpwave_analysis_settings.estimator["mean"] = []
settings.sharpwave_analysis_settings.sharpwave_features.enable_all()
for sw_feature in settings.sharpwave_analysis_settings.sharpwave_features.list_all():
    settings.sharpwave_analysis_settings.estimator["mean"].append(sw_feature)

channels = nm.utils.set_channels(
    ch_names=raw.ch_names,
    ch_types=raw.get_channel_types(),
    reference="default",
    bads=raw.info["bads"],
    new_names="default",
    used_types=("ecog", "dbs", "seeg"),
    target_keywords=["MOV_RIGHT"],
)

stream = nm.Stream(
    sfreq=sfreq,
    channels=channels,
    settings=settings,
    line_noise=line_noise,
    coord_list=coord_list,
    coord_names=coord_names,
    verbose=False,
)
sw_analyzer = cast(
    SharpwaveAnalyzer, stream.data_processor.features.get_feature("sharpwave_analysis")
)

# %%
# The plotted example time series, visualized on a short time scale, shows the relation of identified peaks, troughs, and estimated features:
data_plt = data[5, 1000:4000]

filtered_dat = fftconvolve(data_plt, sw_analyzer.list_filter[0][1], mode="same")

troughs = signal.find_peaks(-filtered_dat, distance=10)[0]
peaks = signal.find_peaks(filtered_dat, distance=5)[0]

sw_results = sw_analyzer.analyze_waveform(filtered_dat)

WIDTH = BAR_WIDTH = 4
BAR_OFFSET = 50
OFFSET_TIME_SERIES = -100
SCALE_TIMESERIES = 1

hue_colors = sb.color_palette("viridis_r", 6)

plt.figure(figsize=(5, 3), dpi=300)
plt.plot(
    OFFSET_TIME_SERIES + data_plt,
    color="gray",
    linewidth=0.5,
    alpha=0.5,
    label="original ECoG data",
)
plt.plot(
    OFFSET_TIME_SERIES + filtered_dat * SCALE_TIMESERIES,
    linewidth=0.5,
    color="black",
    label="[5-30]Hz filtered data",
)

plt.plot(
    peaks,
    OFFSET_TIME_SERIES + filtered_dat[peaks] * SCALE_TIMESERIES,
    "x",
    label="peaks",
    markersize=3,
    color="darkgray",
)
plt.plot(
    troughs,
    OFFSET_TIME_SERIES + filtered_dat[troughs] * SCALE_TIMESERIES,
    "x",
    label="troughs",
    markersize=3,
    color="lightgray",
)

plt.bar(
    troughs + BAR_WIDTH,
    np.array(sw_results["prominence"]) * 4,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[0],
    label="Prominence",
    alpha=0.5,
)
plt.bar(
    troughs + BAR_WIDTH * 2,
    -np.array(sw_results["sharpness"]) * 6,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[1],
    label="Sharpness",
    alpha=0.5,
)
plt.bar(
    troughs + BAR_WIDTH * 3,
    np.array(sw_results["interval"]) * 5,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[2],
    label="Interval",
    alpha=0.5,
)
plt.bar(
    troughs + BAR_WIDTH * 4,
    np.array(sw_results["rise_time"]) * 5,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[3],
    label="Rise time",
    alpha=0.5,
)

plt.xticks(
    np.arange(0, data_plt.shape[0], 200),
    np.round(np.arange(0, int(data_plt.shape[0] / 1000), 0.2), 2),
)
plt.xlabel("Time [s]")
plt.title("Temporal waveform shape features")
plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))
plt.ylim(-550, 700)
plt.xlim(0, 200)
plt.ylabel("a.u.")
plt.tight_layout()

# %%
# See in the following example a time series example, that is aligned to movement. With movement onset the prominence, sharpness, and interval features are reduced:

plt.figure(figsize=(8, 5), dpi=300)
plt.plot(
    OFFSET_TIME_SERIES + data_plt,
    color="gray",
    linewidth=0.5,
    alpha=0.5,
    label="original ECoG data",
)
plt.plot(
    OFFSET_TIME_SERIES + filtered_dat * SCALE_TIMESERIES,
    linewidth=0.5,
    color="black",
    label="[5-30]Hz filtered data",
)

plt.plot(
    peaks,
    OFFSET_TIME_SERIES + filtered_dat[peaks] * SCALE_TIMESERIES,
    "x",
    label="peaks",
    markersize=3,
    color="darkgray",
)
plt.plot(
    troughs,
    OFFSET_TIME_SERIES + filtered_dat[troughs] * SCALE_TIMESERIES,
    "x",
    label="troughs",
    markersize=3,
    color="lightgray",
)

plt.bar(
    troughs + BAR_WIDTH,
    np.array(sw_results["prominence"]) * 4,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[0],
    label="Prominence",
    alpha=0.5,
)
plt.bar(
    troughs + BAR_WIDTH * 2,
    -np.array(sw_results["sharpness"]) * 6,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[1],
    label="Sharpness",
    alpha=0.5,
)
plt.bar(
    troughs + BAR_WIDTH * 3,
    np.array(sw_results["interval"]) * 5,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[2],
    label="Interval",
    alpha=0.5,
)
plt.bar(
    troughs + BAR_WIDTH * 4,
    np.array(sw_results["rise_time"]) * 5,
    bottom=BAR_OFFSET,
    width=WIDTH,
    color=hue_colors[3],
    label="Rise time",
    alpha=0.5,
)

plt.axvline(x=1500, label="Movement start", color="red")

# plt.xticks(np.arange(0, 2000, 200), np.round(np.arange(0, 2, 0.2), 2))
plt.xticks(
    np.arange(0, data_plt.shape[0], 200),
    np.round(np.arange(0, int(data_plt.shape[0] / 1000), 0.2), 2),
)
plt.xlabel("Time [s]")
plt.title("Temporal waveform shape features")
plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))
plt.ylim(-450, 400)
plt.ylabel("a.u.")
plt.tight_layout()

# %%
# In the *sharpwave_analysis_settings* the *estimator* keyword further specifies which statistic is computed based on the individual
# features in one batch. The "global" setting *segment_length_features_ms* specifies the time duration for feature computation.
# Since there can be a different number of identified waveform shape features for different batches (i.e. different number of peaks/troughs),
# taking a statistical measure (e.g. the maximum or mean) will be necessary for feature comparison.

# %%
# Example time series computation for movement decoding
# -----------------------------------------------------
# We will now read the ECoG example/data and investigate if samples differ across movement states. Therefore we compute features and enable the default *sharpwave* features.

settings = NMSettings.get_default().reset()

settings.features.sharpwave_analysis = True
settings.sharpwave_analysis_settings.filter_ranges_hz = [[5, 80]]

channels["used"] = 0  # set only two ECoG channels for faster computation to true
channels.loc[[3, 8], "used"] = 1

stream = nm.Stream(
    sfreq=sfreq,
    channels=channels,
    settings=settings,
    line_noise=line_noise,
    coord_list=coord_list,
    coord_names=coord_names,
    verbose=True,
)

df_features = stream.run(data=data[:, :30000], save_csv=True)

# %%
# We can then plot two exemplary features, prominence and interval, and see that the movement amplitude can be clustered with those two features alone:

plt.figure(figsize=(5, 3), dpi=300)
print(df_features.columns)
plt.scatter(
    df_features["ECOG_RIGHT_0_avgref_Sharpwave_Max_prominence_range_5_80"],
    df_features["ECOG_RIGHT_5_avgref_Sharpwave_Mean_interval_range_5_80"],
    c=df_features.MOV_RIGHT,
    alpha=0.8,
    s=30,
)
cbar = plt.colorbar()
cbar.set_label("Movement amplitude")
plt.xlabel("Prominence a.u.")
plt.ylabel("Interval a.u.")
plt.title("Temporal features predict movement amplitude")
plt.tight_layout()



import numpy as np
from collections.abc import Sequence


class MNEFilter:
    """mne.filter wrapper

    This class stores for given frequency band ranges the filter
    coefficients with length "filter_len".
    The filters can then be used sequentially for band power estimation with
    apply_filter().
    Note that this filter can be a bandpass, bandstop, lowpass, or highpass filter
    depending on the frequency ranges given (see further details in mne.filter.create_filter).

    Parameters
    ----------
    f_ranges : list[tuple[float | None, float | None]]
    sfreq : float
        Sampling frequency.
    filter_length : str, optional
        Filter length. Human readable (e.g. "1000ms", "1s"), by default "999ms"
    l_trans_bandwidth : float | str, optional
        Length of the lower transition band or "auto", by default 4
    h_trans_bandwidth : float | str, optional
        Length of the higher transition band or "auto", by default 4
    verbose : bool | None, optional
        Verbosity level, by default None

    Attributes
    ----------
    filter_bank: np.ndarray shape (n,)
        Factor to upsample by.
    """

    def __init__(
        self,
        f_ranges: Sequence[tuple[float | None, float | None]],
        sfreq: float,
        filter_length: str | float = "999ms",
        l_trans_bandwidth: float | str = 4,
        h_trans_bandwidth: float | str = 4,
        verbose: bool | int | str | None = None,
    ) -> None:
        from mne.filter import create_filter

        filter_bank = []
        # mne create_filter function only accepts str and int for filter_length
        if isinstance(filter_length, float):
            filter_length = int(filter_length)

        for f_range in f_ranges:
            try:
                filt = create_filter(
                    None,
                    sfreq,
                    l_freq=f_range[0],
                    h_freq=f_range[1],
                    fir_design="firwin",
                    l_trans_bandwidth=l_trans_bandwidth,  # type: ignore
                    h_trans_bandwidth=h_trans_bandwidth,  # type: ignore
                    filter_length=filter_length,  # type: ignore
                    verbose=verbose,
                )
            except ValueError:
                filt = create_filter(
                    None,
                    sfreq,
                    l_freq=f_range[0],
                    h_freq=f_range[1],
                    fir_design="firwin",
                    verbose=verbose,
                    # filter_length=filter_length,
                )
            filter_bank.append(filt)

        self.num_filters = len(filter_bank)
        self.filter_bank = np.vstack(filter_bank)

        self.filters: np.ndarray
        self.num_channels = -1

    def filter_data(self, data: np.ndarray) -> np.ndarray:
        """Apply previously calculated (bandpass) filters to data.

        Parameters
        ----------
        data : np.ndarray (n_samples, ) or (n_channels, n_samples)
            Data to be filtered
        filter_bank : np.ndarray, shape (n_fbands, filter_len)
            Output of calc_bandpass_filters.

        Returns
        -------
        np.ndarray, shape (n_channels, n_fbands, n_samples)
            Filtered data.

        """
        from scipy.signal import fftconvolve

        if data.ndim > 2:
            raise ValueError(
                f"Data must have one or two dimensions. Got:"
                f" {data.ndim} dimensions."
            )
        if data.ndim == 1:
            data = np.expand_dims(data, axis=0)

        if self.num_channels == -1:
            self.num_channels = data.shape[0]
            self.filters = np.tile(
                self.filter_bank[None, :, :], (self.num_channels, 1, 1)
            )

        data_tiled = np.tile(data[:, None, :], (1, self.num_filters, 1))

        filtered = fftconvolve(data_tiled, self.filters, axes=2, mode="same")

        # ensure here that the output dimension matches the input dimension
        if data.shape[1] != filtered.shape[-1]:
            # select the middle part of the filtered data
            middle_index = filtered.shape[-1] // 2
            filtered = filtered[
                :,
                :,
                middle_index - data.shape[1] // 2 : middle_index + data.shape[1] // 2,
            ]

        return filtered



import random
import copy

import matplotlib.pyplot as plt

# from numba import njit
import numpy as np
import pandas as pd
import scipy.stats as scipy_stats


def fitlm_kfold(x, y, kfold_splits=5):
    from sklearn.linear_model import LinearRegression
    from sklearn.model_selection import KFold

    model = LinearRegression()
    if isinstance(x, type(np.array([]))) or isinstance(x, type([])):
        x = pd.DataFrame(x)
    if isinstance(y, type(np.array([]))) or isinstance(y, type([])):
        y = pd.DataFrame(y)
    scores, coeffs = [], np.zeros(x.shape[1])
    kfold = KFold(n_splits=kfold_splits, shuffle=True, random_state=42)
    for i, (train, test) in enumerate(kfold.split(x, y)):
        model.fit(x.iloc[train, :], y.iloc[train, :])
        score = model.score(x.iloc[test, :], y.iloc[test, :])
        scores.append(score)
        coeffs = np.vstack((coeffs, model.coef_))
    coeffs = list(np.delete(coeffs, 0))
    return scores, coeffs, model, ["scores", "coeffs", "model"]


def zscore(data):
    return (data - data.mean()) / data.std()


def permutationTestSpearmansRho(x, y, plot_distr=True, x_unit=None, p=5000):
    """
    Calculate permutation test for multiple repetitions of Spearmans Rho
    https://towardsdatascience.com/how-to-assess-statistical-significance-in-your-data-with-permutation-tests-8bb925b2113d

    x (np array) : first distibution e.g. R^2
    y (np array) : second distribution e.g. UPDRS
    plot_distr (boolean) : if True: permutation histplot and ground truth will be
    plotted
    x_unit (str) : histplot xlabel
    p (int): number of permutations

    returns:
    gT (float) : estimated ground truth, here spearman's rho
    p (float) : p value of permutation test
    """

    # compute ground truth difference
    gT = scipy_stats.spearmanr(x, y)[0]
    #
    pV = np.array((x, y))
    # Initialize permutation:
    pD = []
    # Permutation loop:
    args_order = np.arange(0, pV.shape[1], 1)
    args_order_2 = np.arange(0, pV.shape[1], 1)
    for i in range(0, p):
        # Shuffle the data:
        random.shuffle(args_order)
        random.shuffle(args_order_2)
        # Compute permuted absolute difference of your two sampled
        # distributions and store it in pD:
        pD.append(scipy_stats.spearmanr(pV[0, args_order], pV[1, args_order_2])[0])

    # calculate p value
    if gT < 0:
        p_val = len(np.where(pD <= gT)[0]) / p
    else:
        p_val = len(np.where(pD >= gT)[0]) / p

    if plot_distr:
        plt.hist(pD, bins=30, label="permutation results")
        plt.axvline(gT, color="orange", label="ground truth")
        plt.title("ground truth " + x_unit + "=" + str(gT) + " p=" + str(p_val))
        plt.xlabel(x_unit)
        plt.legend()
        plt.show()
    return gT, p_val


def permutationTest(x, y, plot_distr=True, x_unit=None, p=5000):
    """
    Calculate permutation test
    https://towardsdatascience.com/how-to-assess-statistical-significance-in-your-data-with-permutation-tests-8bb925b2113d

    x (np array) : first distr.
    y (np array) : first distr.
    plot_distr (boolean) : if True: plot permutation histplot and ground truth
    x_unit (str) : histplot xlabel
    p (int): number of permutations

    returns:
    gT (float) : estimated ground truth, here absolute difference of
    distribution means
    p (float) : p value of permutation test

    """
    # Compute ground truth difference
    gT = np.abs(np.average(x) - np.average(y))

    pV = np.concatenate((x, y), axis=0)
    pS = copy.copy(pV)
    # Initialize permutation:
    pD = []
    # Permutation loop:
    for i in range(0, p):
        # Shuffle the data:
        random.shuffle(pS)
        # Compute permuted absolute difference of your two sampled
        # distributions and store it in pD:
        pD.append(
            np.abs(
                np.average(pS[0 : int(len(pS) / 2)])
                - np.average(pS[int(len(pS) / 2) :])
            )
        )

    # Calculate p-value
    if gT < 0:
        p_val = len(np.where(pD <= gT)[0]) / p
    else:
        p_val = len(np.where(pD >= gT)[0]) / p

    if plot_distr:
        plt.hist(pD, bins=30, label="permutation results")
        plt.axvline(gT, color="orange", label="ground truth")
        plt.title("ground truth " + x_unit + "=" + str(gT) + " p=" + str(p_val))
        plt.xlabel(x_unit)
        plt.legend()
        plt.show()
    return gT, p_val


def permutationTest_relative(x, y, plot_distr=True, x_unit=None, p=5000):
    """
    Calculate permutation test
    https://towardsdatascience.com/how-to-assess-statistical-significance-in-your-data-with-permutation-tests-8bb925b2113d

    x (np array) : first distr.
    y (np array) : first distr.
    plot_distr (boolean) : if True: plot permutation histplot and ground truth
    x_unit (str) : histplot xlabel
    p (int): number of permutations

    returns:
    gT (float) : estimated ground truth, here absolute difference of
    distribution means
    p (float) : p value of permutation test

    """
    gT = np.abs(np.average(x) - np.average(y))
    pD = []
    for i in range(0, p):
        l_ = []
        for i in range(x.shape[0]):
            if random.randint(0, 1) == 1:
                l_.append((x[i], y[i]))
            else:
                l_.append((y[i], x[i]))
        pD.append(
            np.abs(np.average(np.array(l_)[:, 0]) - np.average(np.array(l_)[:, 1]))
        )
    if gT < 0:
        p_val = len(np.where(pD <= gT)[0]) / p
    else:
        p_val = len(np.where(pD >= gT)[0]) / p

    if plot_distr:
        plt.hist(pD, bins=30, label="permutation results")
        plt.axvline(gT, color="orange", label="ground truth")
        plt.title("ground truth " + x_unit + "=" + str(gT) + " p=" + str(p_val))
        plt.xlabel(x_unit)
        plt.legend()
        plt.show()

    return gT, p_val


# @njit
def permutation_numba_onesample(x, y, n_perm, two_tailed=True):
    """Perform permutation test with one-sample distribution.

    Parameters
    ----------
    x : array_like
        First distribution
    y : int or float
        Baseline against which to check for statistical significane
    n_perm : int
        Number of permutations
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-sampled permutation
        test, else True
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-tailed permutation
        test, else True

    Returns
    -------
    float
        Estimated difference of distribution from baseline
    float
        P-value of permutation test
    """
    if two_tailed:
        zeroed = x - y
        z = np.abs(np.mean(zeroed))
        p = np.empty(n_perm)
        # Run the simulation n_perm times
        for i in np.arange(n_perm):
            sign = np.random.choice(a=np.array([-1.0, 1.0]), size=len(x), replace=True)
            p[i] = np.abs(np.mean(zeroed * sign))
    else:
        zeroed = x - y
        z = np.mean(zeroed)
        p = np.empty(n_perm)
        # Run the simulation n_perm times
        for i in np.arange(n_perm):
            sign = np.random.choice(a=np.array([-1.0, 1.0]), size=len(x), replace=True)
            p[i] = np.mean(zeroed * sign)
        # Return p-value
    return z, (np.sum(p >= z)) / n_perm


# @njit
def permutation_numba_twosample(x, y, n_perm, two_tailed=True):
    """Perform permutation test.

    Parameters
    ----------
    x : array_like
        First distribution
    y : array_like
        Second distribution
    n_perm : int
        Number of permutations
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-sampled permutation
        test, else True
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-tailed permutation
        test, else True

    Returns
    -------
    float
        Estimated difference of distribution means
    float
        P-value of permutation test
    """
    if two_tailed:
        z = np.abs(np.mean(x) - np.mean(y))
        pS = np.concatenate((x, y), axis=0)
        half = int(len(pS) / 2)
        p = np.empty(n_perm)
        # Run the simulation n_perm times
        for i in np.arange(0, n_perm):
            # Shuffle the data
            np.random.shuffle(pS)
            # Compute permuted absolute difference of the two sampled
            # distributions
            p[i] = np.abs(np.mean(pS[:half]) - np.mean(pS[half:]))
    else:
        z = np.mean(x) - np.mean(y)
        pS = np.concatenate((x, y), axis=0)
        half = int(len(pS) / 2)
        p = np.empty(n_perm)
        # Run the simulation n_perm times
        for i in np.arange(0, n_perm):
            # Shuffle the data
            np.random.shuffle(pS)
            # Compute permuted absolute difference of the two sampled
            # distributions
            p[i] = np.mean(pS[:half]) - np.mean(pS[half:])
    return z, (np.sum(p >= z)) / n_perm


def cluster_wise_p_val_correction(p_arr, p_sig=0.05, num_permutations=10000):
    """Obtain cluster-wise corrected p values.

    Based on: https://github.com/neuromodulation/wjn_toolbox/blob/4745557040ad26f3b8498ca5d0c5d5dece2d3ba1/mypcluster.m
    https://garstats.wordpress.com/2018/09/06/cluster/

    Arguments
    ---------
    p_arr (np.array) : ndim, can be time series or image
    p_sig (float) : significance level
    num_permutations (int) : no. of random permutations of cluster comparisons

    Returns
    -------
    p (float) : significance level of highest cluster
    p_min_index : indices of significant samples
    """
    from scipy.ndimage import label as measure_label

    labels, num_clusters = measure_label(p_arr <= p_sig)

    # loop through clusters of p_val series or image
    index_cluster = {}
    p_cluster_sum = np.zeros(num_clusters)
    for cluster_i in np.arange(num_clusters):
        # first cluster is assigned to be 1 from measure.label
        index_cluster[cluster_i] = np.where(labels == cluster_i + 1)[0]
        p_cluster_sum[cluster_i] = np.sum(np.array(1 - p_arr)[index_cluster[cluster_i]])
    # p_min corresponds to the most unlikely cluster
    p_min = np.max(p_cluster_sum)

    p_min_index = index_cluster[np.argmax(p_cluster_sum)]

    # loop through random permutation cycles
    r_per_arr = np.zeros(num_permutations)
    for r in range(num_permutations):
        r_per = np.random.randint(low=0, high=p_arr.shape[0], size=p_arr.shape[0])

        labels, num_clusters = measure_label(p_arr[r_per] <= p_sig, return_num=True)

        index_cluster = {}
        if num_clusters == 0:
            r_per_arr[r] = 0
        else:
            p_cluster_sum = np.zeros(num_clusters)
            for cluster_i in np.arange(num_clusters):
                index_cluster[cluster_i] = np.where(labels == cluster_i + 1)[
                    0
                ]  # first cluster is assigned to be 1 from measure.label
                p_cluster_sum[cluster_i] = np.sum(
                    np.array(1 - p_arr[r_per])[index_cluster[cluster_i]]
                )
            # corresponds to the most unlikely cluster
            r_per_arr[r] = np.max(p_cluster_sum)

    sorted_r = np.sort(r_per_arr)

    def find_arg_nearest(array, value):
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return idx

    p = 1 - find_arg_nearest(sorted_r, p_min) / num_permutations

    return p, p_min_index


# @njit
def cluster_wise_p_val_correction_numba(p_arr, p_sig, n_perm):
    """Calculate significant clusters and their corresponding p-values.

    Based on:
    https://github.com/neuromodulation/wjn_toolbox/blob/4745557040ad26f3b8498ca5d0c5d5dece2d3ba1/mypcluster.m
    https://garstats.wordpress.com/2018/09/06/cluster/

    Arguments
    ---------
    p_arr :  array-like
        Array of p-values. WARNING: MUST be one-dimensional
    p_sig : float
        Significance level
    n_perm : int
        No. of random permutations for building cluster null-distribution

    Returns
    -------
    p : list of floats
        List of p-values for each cluster
    p_min_index : list of numpy array
        List of indices of each significant cluster
    """

    def cluster(iterable):
        """Cluster 1-D array of boolean values.

        Parameters
        ----------
        iterable : array-like of bool
            Array to be clustered.

        Returns
        -------
        cluster_labels : np.ndarray
            Array of shape (len(iterable), 1), where each value indicates the
            number of the cluster. Values are 0 if the item does not belong to
            a cluster
        cluster_count : int
            Number of detected cluster. Corresponds to the highest value in
            cluster_labels
        """
        cluster_labels = np.zeros((len(iterable), 1))
        cluster_count = 0
        cluster_len = 0
        for idx, item in enumerate(iterable):
            if item:
                cluster_labels[idx] = cluster_count + 1
                cluster_len += 1
            elif cluster_len == 0:
                pass
            else:
                cluster_len = 0
                cluster_count += 1
        if cluster_len >= 1:
            cluster_count += 1
        return cluster_labels, cluster_count

    def calculate_null_distribution(p_arr_, p_sig_, n_perm_):
        """Calculate null distribution of clusters.

        Parameters
        ----------
        p_arr_ :  numpy array
            Array of p-values
        p_sig_ : float
            Significance level (p-value)
        n_perm_ : int
            No. of random permutations

        Returns
        -------
        r_per_arr : numpy array
            Null distribution of shape (n_perm_)
        """
        # loop through random permutation cycles
        r_per_arr = np.zeros(n_perm_)
        for r in range(n_perm_):
            r_per = np.random.randint(low=0, high=p_arr_.shape[0], size=p_arr_.shape[0])
            labels_, n_clusters = cluster(p_arr_[r_per] <= p_sig_)

            cluster_ind = {}
            if n_clusters == 0:
                r_per_arr[r] = 0
            else:
                p_sum = np.zeros(n_clusters)
                for ind in range(n_clusters):
                    cluster_ind[ind] = np.where(labels_ == ind + 1)[0]
                    p_sum[ind] = np.sum(np.asarray(1 - p_arr_[r_per])[cluster_ind[ind]])
                r_per_arr[r] = np.max(p_sum)
        return r_per_arr

    labels, num_clusters = cluster(p_arr <= p_sig)

    null_distr = calculate_null_distribution(p_arr, p_sig, n_perm)
    # Loop through clusters of p_val series or image
    clusters = []
    p_vals = [np.float64(x) for x in range(0)]
    # Cluster labels start at 1
    for cluster_i in range(num_clusters):
        index_cluster = np.where(labels == cluster_i + 1)[0]
        p_cluster_sum = np.sum(np.asarray(1 - p_arr)[index_cluster])
        p_val = 1 - np.sum(p_cluster_sum >= null_distr) / n_perm
        if p_val <= p_sig:
            clusters.append(index_cluster)
            p_vals.append(p_val)

    return p_vals, clusters



import numpy as np
from collections.abc import Iterable

from typing import TYPE_CHECKING

from py_neuromodulation.utils.types import NMFeature, BoolSelector, NMBaseModel

from pydantic import field_validator

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings


class NoldsFeatures(BoolSelector):
    sample_entropy: bool = False
    correlation_dimension: bool = False
    lyapunov_exponent: bool = True
    hurst_exponent: bool = False
    detrended_fluctuation_analysis: bool = False


class NoldsSettings(NMBaseModel):
    raw: bool = True
    frequency_bands: list[str] = ["low_beta"]
    features: NoldsFeatures = NoldsFeatures()

    @field_validator("frequency_bands")
    def fbands_spaces_to_underscores(cls, frequency_bands):
        return [f.replace(" ", "_") for f in frequency_bands]


class Nolds(NMFeature):
    def __init__(
        self, settings: "NMSettings", ch_names: Iterable[str], sfreq: float
    ) -> None:
        self.settings = settings.nolds_settings
        self.ch_names = ch_names

        if len(self.settings.frequency_bands) > 0:
            from py_neuromodulation.features.bandpower import BandPower

            self.bp_filter = BandPower(settings, ch_names, sfreq, use_kf=False)

        # Check if the selected frequency bands are defined in the global settings
        for fb in settings.nolds_settings.frequency_bands:
            assert (
                fb in settings.frequency_ranges_hz
            ), f"{fb} selected in nolds_features, but not defined in s['frequency_ranges_hz']"

    def calc_feature(self, data: np.ndarray) -> dict:
        feature_results = {}
        data = np.nan_to_num(data)
        if self.settings.raw:
            feature_results = self.calc_nolds(data, feature_results)
        if len(self.settings.frequency_bands) > 0:
            data_filt = self.bp_filter.bandpass_filter.filter_data(data)

            for f_band_idx, f_band in enumerate(self.settings.frequency_bands):
                # filter data now for a specific fband and pass to calc_nolds
                feature_results = self.calc_nolds(
                    data_filt[:, f_band_idx, :], feature_results, f_band
                )  # ch, bands, samples
        return feature_results

    def calc_nolds(
        self, data: np.ndarray, feature_results: dict, data_str: str = "raw"
    ) -> dict:
        for ch_idx, ch_name in enumerate(self.ch_names):
            for f_name in self.settings.features.get_enabled():
                feature_results[f"{ch_name}_nolds_{f_name}_{data_str}"] = (
                    self.calc_nolds_feature(f_name, data[ch_idx, :])
                    if data[ch_idx, :].sum()
                    else 0
                )

        return feature_results

    @staticmethod
    def calc_nolds_feature(f_name: str, dat: np.ndarray):
        import nolds

        match f_name:
            case "sample_entropy":
                return nolds.sampen(dat)
            case "correlation_dimension":
                return nolds.corr_dim(dat, emb_dim=2)
            case "lyapunov_exponent":
                return nolds.lyap_r(dat)
            case "hurst_exponent":
                return nolds.hurst_rs(dat)
            case "detrended_fluctuation_analysis":
                return nolds.dfa(dat)
            case _:
                raise ValueError(f"Invalid nolds feature name: {f_name}")



import numpy as np
from numpy.testing import assert_allclose
import pytest

import py_neuromodulation as nm
from py_neuromodulation.processing import ReReferencer


def test_rereference_not_used_channels_no_reref(setup_databatch):
    ch_names, ch_types, bads, data_batch = setup_databatch

    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=bads,
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=("MOV_RIGHT",),
    )

    re_referencer = ReReferencer(1, channels)

    # select here data that will is selected, this operation takes place in the nm_run_analysis
    data_used = data_batch[channels["used"] == 1]

    ref_dat = re_referencer.process(data_used)

    for no_ref_idx in np.where((channels.rereference == "None") & channels.used == 1)[
        0
    ]:
        assert_allclose(ref_dat[no_ref_idx, :], data_batch[no_ref_idx, :])


def test_rereference_car(setup_databatch):
    ch_names, ch_types, bads, data_batch = setup_databatch

    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=bads,
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=("MOV_RIGHT",),
    )

    re_referencer = ReReferencer(1, channels)

    data_used = data_batch[channels["used"] == 1]

    ref_dat = re_referencer.process(data_used)

    for ecog_ch_idx in np.where(
        (channels["type"] == "ecog") & (channels.rereference == "average")
    )[0]:
        assert_allclose(
            ref_dat[ecog_ch_idx, :],
            data_batch[ecog_ch_idx, :]
            - data_batch[
                (channels["type"] == "ecog") & (channels.index != ecog_ch_idx)
            ].mean(axis=0),
        )


def test_rereference_bp(setup_databatch):
    ch_names, ch_types, bads, data_batch = setup_databatch

    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=bads,
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=("MOV_RIGHT",),
    )

    re_referencer = ReReferencer(1, channels)

    data_used = data_batch[channels["used"] == 1]

    ref_dat = re_referencer.process(data_used)

    for bp_reref_idx in [
        ch_idx
        for ch_idx, ch in enumerate(channels.rereference)
        if ch in list(channels.name)
    ]:
        # bp_reref_idx is the channel index of the rereference anode
        # referenced_bp_channel is the channel index which is the rereference cathode
        referenced_bp_channel = np.where(
            channels.iloc[bp_reref_idx]["rereference"] == channels.name
        )[0][0]
        assert_allclose(
            ref_dat[bp_reref_idx, :],
            data_batch[bp_reref_idx, :] - data_batch[referenced_bp_channel, :],
        )


def test_rereference_wrong_rererference_column_name(setup_databatch):
    ch_names, ch_types, bads, data_batch = setup_databatch

    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=bads,
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=("SQUARED_ROTATION",),
    )

    channels.loc[0, "rereference"] = "hallo"
    with pytest.raises(Exception):
        ReReferencer(1, channels)


def test_rereference_muliple_channels(setup_databatch):
    ch_names, ch_types, bads, data_batch = setup_databatch

    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=bads,
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=("MOV_RIGHT",),
    )

    channels.loc[0, "rereference"] = "LFP_RIGHT_1&LFP_RIGHT_2"

    re_referencer = ReReferencer(1, channels)

    data_used = data_batch[channels["used"] == 1]

    ref_dat = re_referencer.process(data_used)

    assert_allclose(
        ref_dat[0, :],
        data_batch[0, :] - (data_batch[1, :] + data_batch[2, :]) / 2,
    )


def test_rereference_same_channel(setup_databatch):
    ch_names, ch_types, bads, data_batch = setup_databatch

    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=bads,
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=("MOV_RIGHT",),
    )

    channels.loc[0, "rereference"] = channels.loc[0, "name"]

    with pytest.raises(Exception):
        ReReferencer(1, channels)



"""Module for generic and offline data streams."""

import time
from typing import TYPE_CHECKING, Any
from collections.abc import Iterator
import numpy as np
from pathlib import Path

import py_neuromodulation as nm

from py_neuromodulation.stream.data_processor import DataProcessor
from py_neuromodulation.utils.types import _PathLike, FEATURE_NAME
from py_neuromodulation.utils.file_writer import MsgPackFileWriter
from py_neuromodulation.stream.settings import NMSettings
from py_neuromodulation.analysis.decode import RealTimeDecoder
from py_neuromodulation.stream.backend_interface import StreamBackendInterface

if TYPE_CHECKING:
    import pandas as pd


class Stream:
    """_GenericStream base class.
    This class can be inherited for different types of offline streams

    Parameters
    ----------
    nm_stream_abc : stream_abc.NMStream
    """

    def __init__(
        self,
        sfreq: float,
        channels: "pd.DataFrame | _PathLike | None" = None,
        data: "np.ndarray | pd.DataFrame | None" = None,
        settings: NMSettings | _PathLike | None = None,
        line_noise: float | None = 50,
        sampling_rate_features_hz: float | None = None,
        path_grids: _PathLike | None = None,
        coord_names: list | None = None,
        coord_list: list | None = None,
        verbose: bool = True,
    ) -> None:
        """Stream initialization

        Parameters
        ----------
        sfreq : float
            sampling frequency of data in Hertz
        channels : pd.DataFrame | _PathLike
            parametrization of channels (see define_channels.py for initialization)
        data : np.ndarray | pd.DataFrame | None, optional
            data to be streamed with shape (n_channels, n_time), by default None
        settings : NMSettings | _PathLike | None, optional
            Initialized settings.NMSettings object, by default the py_neuromodulation/settings.yaml are read
            and passed into a settings object
        line_noise : float | None, optional
            line noise, by default 50
        sampling_rate_features_hz : float | None, optional
            feature sampling rate, by default None
        path_grids : _PathLike | None, optional
            path to grid_cortex.tsv and/or gird_subcortex.tsv, by default Non
        coord_names : list | None, optional
            coordinate name in the form [coord_1_name, coord_2_name, etc], by default None
        coord_list : list | None, optional
            coordinates in the form [[coord_1_x, coord_1_y, coord_1_z], [coord_2_x, coord_2_y, coord_2_z],], by default None
        verbose : bool, optional
            print out stream computation time information, by default True
        """
        # This is calling NMSettings.validate() which is making a copy
        self.settings: NMSettings = NMSettings.load(settings)

        if channels is None and data is not None:
            channels = nm.utils.channels.get_default_channels_from_data(data)

        if channels is not None:
            self.channels = nm.io.load_channels(channels)

        if self.channels.query("used == 1 and target == 0").shape[0] == 0:
            raise ValueError(
                "No channels selected for analysis that have column 'used' = 1 and 'target' = 0. Please check your channels"
            )

        if channels is None and data is None:
            raise ValueError("Either `channels` or `data` must be passed to `Stream`.")

        # If features that use frequency ranges are on, test them against nyquist frequency
        use_freq_ranges: list[FEATURE_NAME] = [
            "bandpass_filter",
            "stft",
            "fft",
            "welch",
            "bursts",
            "coherence",
            "nolds",
            "bispectrum",
        ]

        need_nyquist_check = any(
            (f in use_freq_ranges for f in self.settings.features.get_enabled())
        )

        if need_nyquist_check:
            assert all(
                fb.frequency_high_hz < sfreq / 2
                for fb in self.settings.frequency_ranges_hz.values()
            ), (
                "If a feature that uses frequency ranges is selected, "
                "the frequency band ranges need to be smaller than the nyquist frequency.\n"
                f"Got sfreq = {sfreq} and fband ranges:\n {self.settings.frequency_ranges_hz}"
            )

        if sampling_rate_features_hz is not None:
            self.settings.sampling_rate_features_hz = sampling_rate_features_hz

        if path_grids is None:
            path_grids = nm.PYNM_DIR

        self.path_grids = path_grids
        self.verbose = verbose
        self.sfreq = sfreq
        self.line_noise = line_noise
        self.coord_names = coord_names
        self.coord_list = coord_list
        self.sess_right = None
        self.projection = None
        self.model = None
        self.is_running = False

        self.data_processor = DataProcessor(
            sfreq=self.sfreq,
            settings=self.settings,
            channels=self.channels,
            path_grids=self.path_grids,
            coord_names=coord_names,
            coord_list=coord_list,
            line_noise=line_noise,
            verbose=self.verbose,
        )

        self.data = data

        self.target_idx_initialized: bool = False

    def _add_target(self, feature_dict: dict, data: np.ndarray) -> None:
        """Add target channels to feature series.

        Parameters
        ----------
        feature_dict : dict
        data : np.ndarray
            Raw data with shape (n_channels, n_samples).
            Channels not usd for feature computation are also included

        Returns
        -------
        dict
            feature dict with target channels added
        """

        if self.channels["target"].sum() > 0:
            if not self.target_idx_initialized:
                self.target_indexes = self.channels[self.channels["target"] == 1].index
                self.target_names = self.channels.loc[
                    self.target_indexes, "name"
                ].to_list()
                self.target_idx_initialized = True

            for target_idx, target_name in zip(self.target_indexes, self.target_names):
                feature_dict[target_name] = data[target_idx, -1]

    def _handle_data(self, data: "np.ndarray | pd.DataFrame") -> np.ndarray:
        names_expected = self.channels["name"].to_list()

        if isinstance(data, np.ndarray):
            if not len(names_expected) == data.shape[0]:
                raise ValueError(
                    "If data is passed as an array, the first dimension must"
                    " match the number of channel names in `channels`.\n"
                    f" Number of data channels (data.shape[0]): {data.shape[0]}\n"
                    f' Length of channels["name"]: {len(names_expected)}.'
                )
            return data

        names_data = data.columns.to_list()
        if not (
            len(names_expected) == len(names_data)
            and sorted(names_expected) == sorted(names_data)
        ):
            raise ValueError(
                "If data is passed as a DataFrame, the"
                "column names must match the channel names in `channels`.\n"
                f"Input dataframe column names: {names_data}\n"
                f'Expected (from channels["name"]): : {names_expected}.'
            )
        return data.to_numpy().transpose()

    def run(
        self,
        data: "np.ndarray | pd.DataFrame | None" = None,
        out_dir: _PathLike = "",
        experiment_name: str = "sub",
        is_stream_lsl: bool = False,
        stream_lsl_name: str | None = None,
        save_csv: bool = True,
        save_interval: int = 10,
        return_df: bool = True,
        simulate_real_time: bool = False,
        decoder: RealTimeDecoder | None = None,
        backend_interface: StreamBackendInterface | None = None,
    ) -> "pd.DataFrame":
        self.is_stream_lsl = is_stream_lsl
        self.stream_lsl_name = stream_lsl_name
        self.save_csv = save_csv
        self.save_interval = save_interval
        self.return_df = return_df
        self.out_dir = Path.cwd() if not out_dir else Path(out_dir)
        self.experiment_name = experiment_name

        # Validate input data
        if data is not None:
            data = self._handle_data(data)
        elif self.data is not None:
            data = self._handle_data(self.data)
        elif self.data is None and data is None and self.is_stream_lsl is False:
            raise ValueError("No data passed to run function.")

        file_writer = MsgPackFileWriter(name=experiment_name, out_dir=out_dir)

        self.batch_count: int = 0  # Keep track of the number of batches processed

        self.data_processor = DataProcessor(
            sfreq=self.sfreq,
            settings=self.settings,
            channels=self.channels,
            path_grids=self.path_grids,
            coord_names=self.coord_names,
            coord_list=self.coord_list,
            line_noise=self.line_noise,
            verbose=self.verbose,
        )

        nm.logger.log_to_file(out_dir)

        self.generator: Iterator
        if not is_stream_lsl and data is not None:
            from py_neuromodulation.stream.generator import RawDataGenerator

            self.generator = RawDataGenerator(
                data,
                self.sfreq,
                self.settings.sampling_rate_features_hz,
                self.settings.segment_length_features_ms,
            )
            nm.logger.info("Initializing RawDataGenerator")
        else:
            from py_neuromodulation.stream.mnelsl_stream import LSLStream

            self.lsl_stream = LSLStream(
                settings=self.settings, stream_name=stream_lsl_name
            )

            if (
                self.lsl_stream.stream.sinfo is not None
                and self.sfreq != self.lsl_stream.stream.sinfo.sfreq
            ):
                error_msg = (
                    f"Sampling frequency of the lsl-stream ({self.lsl_stream.stream.sinfo.sfreq}) "
                    f"does not match the settings ({self.sfreq})."
                    "The sampling frequency read from the stream will be used"
                )
                nm.logger.warning(error_msg)
                self.sfreq = self.lsl_stream.stream.sinfo.sfreq

            self.generator = self.lsl_stream.get_next_batch()

        prev_batch_end = 0
        for timestamps, data_batch in self.generator:
            self.is_running = True
            if backend_interface:
                # Only simulate real-time if connected to GUI
                if simulate_real_time:
                    time.sleep(1 / self.settings.sampling_rate_features_hz)

                signal = backend_interface.check_control_signals()
                if signal == "stop":
                    break

            if data_batch is None:
                nm.logger.info("Data batch is None, stopping run function")
                break

            nm.logger.info("Processing new data batch")
            feature_dict = self.data_processor.process(data_batch)

            this_batch_end = timestamps[-1]
            batch_length = this_batch_end - prev_batch_end
            nm.logger.debug(
                f"{batch_length:.3f} seconds of new data processed",
            )

            if decoder is not None:
                ch_to_decode = self.channels.query("used == 1").iloc[0]["name"]
                feature_dict = decoder.predict(
                    feature_dict, ch_to_decode, fft_bands_only=True
                )

            feature_dict["time"] = np.ceil(this_batch_end * 1000 + 1)
            prev_batch_end = this_batch_end

            if self.verbose:
                nm.logger.info("Time: %.2f", feature_dict["time"] / 1000)

            self._add_target(feature_dict, data_batch)

            # Push data to file writer
            file_writer.insert_data(feature_dict)

            # Send data to frontend
            if backend_interface:
                nm.logger.debug("stream.run: Sending features to frontend")
                backend_interface.send_features(feature_dict)
                backend_interface.send_raw_data(self._prepare_raw_data_dict(data_batch))

            # Save features to file in intervals
            self.batch_count += 1
            if self.batch_count % self.save_interval == 0:
                file_writer.save()

        file_writer.save()

        if self.save_csv:
            file_writer.save_as_csv(save_all_combined=True)

        feature_df = file_writer.load_all() if self.return_df else {}

        self._save_after_stream()
        self.is_running = False

        return feature_df  # Timon: We could think of returnader instead

    def _prepare_raw_data_dict(self, data_batch: np.ndarray) -> dict[str, Any]:
        """Prepare raw data dictionary for sending through queue"""
        new_time_ms = 1000 / self.settings.sampling_rate_features_hz
        new_samples = int(new_time_ms * self.sfreq / 1000)
        return {
            "raw_data": {
                ch: list(data_batch[i, -new_samples:])
                for i, ch in enumerate(self.channels["name"])
            }
        }

    def plot_raw_signal(
        self,
        sfreq: float | None = None,
        data: np.ndarray | None = None,
        lowpass: float | None = None,
        highpass: float | None = None,
        picks: list | None = None,
        plot_time: bool = True,
        plot_psd: bool = False,
    ) -> None:
        """Use MNE-RawArray Plot to investigate PSD or raw_signal plot.

        Parameters
        ----------
        sfreq : float
            sampling frequency [Hz]
        data : np.ndarray, optional
            data (n_channels, n_times), by default None
        lowpass: float, optional
            cutoff lowpass filter frequency
        highpass: float, optional
            cutoff highpass filter frequency
        picks: list, optional
            list of channels to plot
        plot_time : bool, optional
            mne.io.RawArray.plot(), by default True
        plot_psd : bool, optional
            mne.io.RawArray.plot(), by default False

        Raises
        ------
        ValueError
            raise Exception when no data is passed
        """
        if data is None:
            if self.data is None:
                raise ValueError("No data passed to plot_raw_signal function.")
            else:
                data = (
                    self.data.to_numpy()
                    if isinstance(self.data, pd.DataFrame)
                    else self.data
                )

        if sfreq is None:
            sfreq = self.sfreq

        if self.channels is not None:
            ch_names = self.channels["name"].to_list()
            ch_types = self.channels["type"].to_list()
        else:
            ch_names = [f"ch_{i}" for i in range(data.shape[0])]
            ch_types = ["ecog" for i in range(data.shape[0])]

        from mne import create_info
        from mne.io import RawArray

        info = create_info(ch_names=ch_names, sfreq=sfreq, ch_types=ch_types)  # type: ignore
        raw = RawArray(data, info)

        if picks is not None:
            raw = raw.pick(picks)
        self.raw = raw
        if plot_time:
            raw.plot(highpass=highpass, lowpass=lowpass)
        if plot_psd:
            raw.compute_psd().plot()

    def _save_after_stream(
        self,
    ) -> None:
        """Save settings, nm_channels and sidecar after run"""
        self._save_sidecar()
        self._save_settings()
        self._save_channels()

    def _save_features(
        self,
        feature_arr: "pd.DataFrame",
    ) -> None:
        nm.io.save_features(feature_arr, self.out_dir, self.experiment_name)

    def _save_channels(self) -> None:
        self.data_processor.save_channels(self.out_dir, self.experiment_name)

    def _save_settings(self) -> None:
        self.data_processor.save_settings(self.out_dir, self.experiment_name)

    def _save_sidecar(self) -> None:
        """Save sidecar incduing fs, coords, sess_right to
        out_path_root and subfolder 'folder_name'"""
        additional_args = {"sess_right": self.sess_right}

        self.data_processor.save_sidecar(
            self.out_dir, self.experiment_name, additional_args
        )



"""Module for real-time data normalization."""

import numpy as np
from typing import TYPE_CHECKING, Callable, Literal, get_args

from py_neuromodulation.utils.pydantic_extensions import NMField
from py_neuromodulation.utils.types import (
    NMBaseModel,
    NORM_METHOD,
    NMPreprocessor,
)

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings

NormalizerType = Literal["raw", "feature"]


class NormalizationSettings(NMBaseModel):
    normalization_time_s: float = NMField(30, gt=0, custom_metadata={"unit": "s"})
    normalization_method: NORM_METHOD = NMField(default="zscore")
    clip: float = NMField(default=3, ge=0, custom_metadata={"unit": "a.u."})

    @staticmethod
    def list_normalization_methods() -> list[NORM_METHOD]:
        return list(get_args(NORM_METHOD))

class FeatureNormalizationSettings(NormalizationSettings):    normalize_psd: bool = False

class Normalizer(NMPreprocessor):
    def __init__(
        self,
        sfreq: float,
        settings: "NMSettings",
        type: NormalizerType,
        **kwargs,
    ) -> None:
        self.type = type
        if self.type == "raw":
            self.settings: NormalizationSettings
        else:
            self.settings: FeatureNormalizationSettings

        match self.type:
            case "raw":
                self.settings = settings.raw_normalization_settings.validate()
                self.add_samples = int(sfreq / settings.sampling_rate_features_hz)
            case "feature":
                self.settings = settings.feature_normalization_settings.validate()
                self.add_samples = 0

        # For type = "feature" sfreq = sampling_rate_features_hz
        self.num_samples_normalize = int(self.settings.normalization_time_s * sfreq)

        self.previous: np.ndarray = np.empty((0, 0))  # Default empty array

        self.method = self.settings.normalization_method
        self.using_sklearn = self.method in ["quantile", "power", "robust", "minmax"]

        if self.using_sklearn:
            import sklearn.preprocessing as skpp

            NORM_METHODS_SKLEARN: dict[NORM_METHOD, Callable] = {
                "quantile": lambda: skpp.QuantileTransformer(n_quantiles=300),
                "robust": skpp.RobustScaler,
                "minmax": skpp.MinMaxScaler,
                "power": skpp.PowerTransformer,
            }

            self.normalizer = norm_sklearn(NORM_METHODS_SKLEARN[self.method]())

        else:
            NORM_FUNCTIONS = {
                "mean": norm_mean,
                "median": norm_median,
                "zscore": norm_zscore,
                "zscore-median": norm_zscore_median,
            }
            self.normalizer = NORM_FUNCTIONS[self.method]

    def process(self, data: np.ndarray) -> np.ndarray:
        """Process normalization.
        Note: raw data has to be internally transposed, s.t. raw and features 
        are normalized in the same way.

        Args:
            data (np.ndarray): shape (channels, n_samples)

        Returns:
            np.ndarray: (channels, n_samples)
        """
                    
        if self.previous.size == 0:  # Check if empty
            self.previous = data
            if self.type == "raw":
                self.previous = self.previous.T
            return data
        if self.type == "raw":
            data = data.T
        self.previous = np.vstack((self.previous, data[-self.add_samples :]))

        data = self.normalizer(data, self.previous)

        if self.settings.clip:
            data = data.clip(min=-self.settings.clip, max=self.settings.clip)

        self.previous = self.previous[-self.num_samples_normalize + 1 :]

        data = np.nan_to_num(data)

        return data if self.type != "raw" else data.T


class RawNormalizer(Normalizer):
    def __init__(self, sfreq: float, settings: "NMSettings", **kwargs,) -> None:
        super().__init__(sfreq, settings, "raw", **kwargs)


class FeatureNormalizer(Normalizer):
    def __init__(self, settings: "NMSettings") -> None:
        super().__init__(settings.sampling_rate_features_hz, settings, "feature")


""" Functions to check for NaN's before deciding which Numpy function to call """


def nan_mean(data: np.ndarray, axis: int) -> np.ndarray:
    return (
        np.nanmean(data, axis=axis)
        if np.any(np.isnan(sum(data)))
        else np.mean(data, axis=axis)
    )


def nan_std(data: np.ndarray, axis: int) -> np.ndarray:
    return (
        np.nanstd(data, axis=axis)
        if np.any(np.isnan(sum(data)))
        else np.std(data, axis=axis)
    )


def nan_median(data: np.ndarray, axis: int) -> np.ndarray:
    return (
        np.nanmedian(data, axis=axis)
        if np.any(np.isnan(sum(data)))
        else np.median(data, axis=axis)
    )


def norm_mean(current, previous):
    mean = nan_mean(previous, axis=0)
    return (current - mean) / mean


def norm_median(current, previous):
    median = nan_median(previous, axis=0)
    return (current - median) / median


def norm_zscore(current, previous):
    std = nan_std(previous, axis=0)
    std[std == 0] = 1  # same behavior as sklearn
    return (current - nan_mean(previous, axis=0)) / std


def norm_zscore_median(current, previous):
    std = nan_std(previous, axis=0)
    std[std == 0] = 1  # same behavior as sklearn
    return (current - nan_median(previous, axis=0)) / std


def norm_sklearn(sknormalizer):
    # For the following methods we check for the shape of current
    # when current is a 1D array, then it is the post-processing normalization,
    # and we need to expand, and remove the extra dimension afterwards
    # When current is a 2D array, then it is pre-processing normalization, and
    # there's no need for expanding.

    def sk_normalizer(current, previous):
        return (
            sknormalizer.fit(np.nan_to_num(previous))
            .transform(
                # if post-processing: pad dimensions to 2
                np.reshape(current, (2 - len(current.shape)) * (1,) + current.shape)
            )
            .squeeze()  # if post-processing: remove extra dimension # type: ignore
        )

    return sk_normalizer



import numpy as np
import pandas as pd

import py_neuromodulation as nm


def get_example_settings(test_arr: np.ndarray) -> tuple[nm.NMSettings, pd.DataFrame]:
    settings = nm.NMSettings.get_fast_compute()

    channels = nm.utils.get_default_channels_from_data(test_arr)

    return settings, channels


def test_label_add_single_target():
    sampling_rate_features = 10
    arr_test = np.random.random([2, 5000])
    target_arr = np.arange(0, 5000, 1)
    arr_test[1, :] = target_arr

    settings, channels = get_example_settings(arr_test)

    channels["target"] = [0, 1]
    channels["used"] = [1, 0]
    channels.loc[1, "name"] = "target_ch"

    settings.sampling_rate_features_hz = sampling_rate_features

    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df = stream.run(
        arr_test, out_dir="./test_data", experiment_name="test_label_add_single_target"
    )

    assert df[
        "target_ch"
    ].is_monotonic_increasing, "Not all target values are ascending"

    # check if all "target_ch" values are in the target_arr
    assert all(
        [i in target_arr for i in df["target_ch"]]
    ), "Not all target values are were estimated correctly from the simulated target array"


def test_label_add_multidimensional_target():
    sampling_rate_features = 10
    arr_test = np.random.random([4, 5000])
    target_arr = np.arange(0, 5000, 1)
    target_arr = np.tile(target_arr, (2, 1))
    arr_test[[0, 2], :] = target_arr

    settings, channels = get_example_settings(arr_test)

    channels["target"] = [1, 0, 1, 0]
    channels["used"] = [0, 1, 0, 1]
    channels.loc[0, "name"] = "target_ch_0"
    channels.loc[2, "name"] = "target_ch_1"

    settings.sampling_rate_features_hz = sampling_rate_features

    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df = stream.run(
        arr_test,
        out_dir="./test_data",
        experiment_name="test_label_add_multidimensional_target",
    )

    for target_ch in ["target_ch_0", "target_ch_1"]:
        assert df[
            target_ch
        ].is_monotonic_increasing, "Not all target values are ascending"

        # check if all "target_ch" values are in the target_arr
        assert all(
            [i in target_arr for i in df[target_ch]]
        ), "Not all target values are were estimated correctly from the simulated target array"


def test_label_add_no_target():
    sampling_rate_features = 10
    arr_test = np.random.random([4, 5000])

    settings, channels = get_example_settings(arr_test)

    settings.sampling_rate_features_hz = sampling_rate_features

    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df = stream.run(
        arr_test, out_dir="./test_data", experiment_name="test_label_add_no_target"
    )

    assert all([col.startswith("ch") or col.startswith("time") for col in df.columns])



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
import sys
from pathlib import Path
from sphinx_gallery.sorting import FileNameSortKey

print("CURRENT WORKING DIRECTORY")
print(Path.cwd())

SCRIPT_DIR = Path(__file__).absolute().parent

if SCRIPT_DIR.name == "source":
    # this check is necessary, so we can also run the script from the root directory
    SCRIPT_DIR = SCRIPT_DIR.parent.parent / "py_neuromodulation"
    
print(f"Script Directory to add: {SCRIPT_DIR}")
sys.path.append(str(SCRIPT_DIR))

print(sys.path)

exclude_patterns = ["_build", "_templates"]


# -- Project information -----------------------------------------------------
project = "py_neuromodulation"
copyright = "2021, Timon Merk"
author = "Timon Merk"

source_parsers = {
    ".md": "recommonmark.parser.CommonMarkParser",
}

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_gallery.gen_gallery",
    "sphinx_togglebutton",
    # "nbsphinx",
]

source_suffix = [
    ".rst",
    ".md",
]

autosummary_generate = True

PYDEVD_DISABLE_FILE_VALIDATION = 1

sphinx_gallery_conf = {
    "examples_dirs": "../../examples",
    "gallery_dirs": "auto_examples",
    "within_subsection_order": FileNameSortKey,
}

templates_path = ["_templates"]
exclude_patterns = []


html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_theme_options = {
    "show_nav_level": 4,
    "icon_links": [
        dict(
            name="GitHub",
            url="https://github.com/neuromodulation/py_neuromodulation",
            icon="fa-brands fa-square-github",
        )
    ],
}

# -- Intersphinx configuration -----------------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numba": ("https://numba.readthedocs.io/en/latest", None),
    "mne": ("https://mne.tools/stable", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
}



import numpy as np
from scipy import signal

from py_neuromodulation import NMSettings, Stream
from py_neuromodulation.processing import PreprocessingFilter


def test_preprocessing_within_pipeline(
    setup_default_stream_fast_compute: tuple[np.ndarray, Stream],
):
    data, stream = setup_default_stream_fast_compute

    stream.settings.preprocessing.append("preprocessing_filter")

    stream.settings.preprocessing_filter.bandstop_filter = True
    stream.settings.preprocessing_filter.bandpass_filter = True
    stream.settings.preprocessing_filter.lowpass_filter = True
    stream.settings.preprocessing_filter.highpass_filter = True

    try:
        _ = stream.run(
            data[:, : int(stream.sfreq * 2)],
            out_dir="./test_data",
            experiment_name="test_preprocessing_within_pipeline",
        )
    except Exception as e:
        assert False, f"Error in pipeline including preprocess filtering : {e}"


def test_preprocessing_filter_lowpass():
    data_batch = np.random.random([1, 1000])

    settings = NMSettings.get_default()
    settings.preprocessing.append("preprocessing_filter")
    settings.preprocessing_filter.lowpass_filter = True
    settings.preprocessing_filter.highpass_filter = False
    settings.preprocessing_filter.bandpass_filter = False
    settings.preprocessing_filter.bandstop_filter = False

    settings.preprocessing_filter.lowpass_filter_cutoff_hz = 100

    sfreq = 1000

    preprocessing_filter = PreprocessingFilter(settings, sfreq)
    data_filtered = preprocessing_filter.process(data_batch)

    # compute a scipy signal welch to check if the filter worked
    f, Pxx = signal.welch(data_batch, fs=sfreq, nperseg=1000)
    f, Pxx_f = signal.welch(data_filtered, fs=sfreq, nperseg=1000)

    # check if the power in the frequency range of the lowpass filter is reduced
    assert np.mean(Pxx_f[0, 100:500]) < np.mean(Pxx[0, 100:500])


def test_preprocessing_filter_highpass():
    data_batch = np.random.random([1, 1000])

    settings = NMSettings.get_default()
    settings.preprocessing.append("preprocessing_filter")
    settings.preprocessing_filter.highpass_filter = True
    settings.preprocessing_filter.lowpass_filter = False
    settings.preprocessing_filter.bandpass_filter = False
    settings.preprocessing_filter.bandstop_filter = False

    settings.preprocessing_filter.highpass_filter_cutoff_hz = 100

    sfreq = 1000

    preprocessing_filter = PreprocessingFilter(settings, sfreq)
    data_filtered = preprocessing_filter.process(data_batch)

    # compute a scipy signal welch to check if the filter worked
    f, Pxx = signal.welch(data_batch, fs=sfreq, nperseg=1000)
    f, Pxx_f = signal.welch(data_filtered, fs=sfreq, nperseg=1000)

    # check if the power in the frequency range of the highpass filter is reduced
    assert np.mean(Pxx_f[0, 0:100]) < np.mean(Pxx[0, 0:100])


def test_preprocessing_filter_bandstop():
    data_batch = np.random.random([1, 1000])

    settings = NMSettings.get_default()
    settings.preprocessing.append("preprocessing_filter")
    settings.preprocessing_filter.bandstop_filter = True
    settings.preprocessing_filter.bandpass_filter = False
    settings.preprocessing_filter.lowpass_filter = False
    settings.preprocessing_filter.highpass_filter = False

    settings.preprocessing_filter.bandstop_filter_settings.frequency_low_hz = 100
    settings.preprocessing_filter.bandstop_filter_settings.frequency_high_hz = 160

    sfreq = 1000

    preprocessing_filter = PreprocessingFilter(settings, sfreq)
    data_filtered = preprocessing_filter.process(data_batch)

    # compute a scipy signal welch to check if the filter worked
    f, Pxx = signal.welch(data_batch, fs=sfreq, nperseg=1000)
    f, Pxx_f = signal.welch(data_filtered, fs=sfreq, nperseg=1000)

    # check if the power in the frequency range of the bandstop filter is reduced
    assert np.mean(Pxx_f[0, 100:160]) < np.mean(Pxx[0, 100:160])


def test_preprocessing_filter_bandpass():
    data_batch = np.random.random([1, 1000])

    settings = NMSettings.get_default()
    settings.preprocessing.append("preprocessing_filter")
    settings.preprocessing_filter.bandstop_filter = False
    settings.preprocessing_filter.bandpass_filter = True
    settings.preprocessing_filter.lowpass_filter = False
    settings.preprocessing_filter.highpass_filter = False

    settings.preprocessing_filter.bandpass_filter_settings.frequency_low_hz = 100
    settings.preprocessing_filter.bandpass_filter_settings.frequency_high_hz = 160

    sfreq = 1000

    preprocessing_filter = PreprocessingFilter(settings, sfreq)
    data_filtered = preprocessing_filter.process(data_batch)

    # compute a scipy signal welch to check if the filter worked
    f, Pxx = signal.welch(data_batch, fs=sfreq, nperseg=1000)
    f, Pxx_f = signal.welch(data_filtered, fs=sfreq, nperseg=1000)

    # check if the power in the frequency range of the bandpass filter is reduced
    assert np.mean(Pxx_f[0, 0:100]) < np.mean(Pxx[0, 0:100])
    assert np.mean(Pxx_f[0, 160:500]) < np.mean(Pxx[0, 160:500])



"""
Cebra Decoding with no training Example
======================================

The following example show how to use the Cebra decoding without training.

"""

import os

# load example_cebra_decoding.html
with open(os.path.join("..", "examples", "example_cebra_decoding.html"), "rt") as fh:
    html_data = fh.read()

tmp_dir = os.path.join("..", "docs", "source", "auto_examples")
if os.path.exists(tmp_dir):
    # building the docs with sphinx-gallery
    with open(os.path.join(tmp_dir, "out.html"), "wt") as fh:
        fh.write(html_data)
# set example path for thumbnail
# sphinx_gallery_thumbnail_path = '_static/CEBRA_embedding.png'


# %%
# CEBRA example
# -------------
#
# .. raw:: html
#     :file: out.html



import nibabel as nib
import numpy as np
from matplotlib import pyplot as plt


class NiiToMNI:
    def __init__(
        self,
        PATH_nii_file: str = r"C:\code\RMap_ROI_Estimation\Automated Anatomical Labeling 3 (Rolls 2020).nii",
    ) -> None:
        self.img = nib.Nifti1Image.from_filename(PATH_nii_file)
        self.data = self.img.get_fdata()

    def downsample_nii(
        self,
        resampling_factor: int = 150,
    ):
        # PATH_MNI_TO_ATLAS = r"C:\code\mni_to_atlas\src\mni_to_atlas\atlases\AAL.nii"
        # img_mni_to_atlas = nib.load(PATH_MNI_TO_ATLAS)

        x_dim, y_dim, z_dim = self.data.shape

        # Create arrays of voxel coordinates
        x_coords, y_coords, z_coords = np.meshgrid(
            range(x_dim), range(y_dim), range(z_dim), indexing="ij"
        )

        # Downsample here the voxels --> check lateron if the voxels have non-zero values
        x_c_flatten = x_coords.flatten()[::resampling_factor]
        y_c_flatten = y_coords.flatten()[::resampling_factor]
        z_c_flatten = z_coords.flatten()[::resampling_factor]

        # Combine coordinates into a single array
        voxel_coordinates = np.column_stack(
            (
                x_c_flatten,
                y_c_flatten,
                z_c_flatten,
                np.ones(x_c_flatten.shape[0]),
            )
        )

        aff_m = self.img.affine
        aff_m[0, 0] = 2
        aff_m[0, 3] = -90

        mni_coordinates = np.dot(aff_m, voxel_coordinates.T).T[:, :3]

        return mni_coordinates

    def select_non_zero_voxels(
        self,
        mni_coordinates: np.ndarray,
    ):
        coords = np.hstack((mni_coordinates, np.ones((mni_coordinates.shape[0], 1))))

        # and transform back to get the voxel values
        voxels_downsampled = np.array(
            np.linalg.solve(self.img.affine, coords.T).T
        ).astype(int)[:, :3]

        ival = []
        coord_ = []
        for i in range(voxels_downsampled.shape[0]):
            ival.append(self.data[tuple(voxels_downsampled[i, :])])
            coord_.append(mni_coordinates[i, :])

        # get only voxel values non-zero
        ival_arr = np.array(ival)
        coord_arr = np.array(coord_)
        ival_non_zero = ival_arr[ival != 0]
        coord_non_zero = coord_arr[ival != 0]

        return coord_non_zero, ival_non_zero

    def plot_3d_coordinates(self, coord_non_zero: np.ndarray):
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(
            coord_non_zero[:, 0],
            coord_non_zero[:, 1],
            coord_non_zero[:, 2],
            s=50,
            alpha=0.2,
        )
        plt.show()


if __name__ == "__main__":
    nii_to_mni = NiiToMNI(
        PATH_nii_file=r"C:\code\py_neuromodulation\ConnectivityDecoding\Automated Anatomical Labeling 3 (Rolls 2020).nii"
    )
    mni_coordinates = nii_to_mni.downsample_nii(resampling_factor=150)
    coord_non_zero, ival_non_zero = nii_to_mni.select_non_zero_voxels(mni_coordinates)
